"""Block Neural Autoregressive bijection implementation."""

from collections.abc import Callable
from math import prod
from typing import ClassVar

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jr
from jax import random
from jaxtyping import PRNGKeyArray

from flowjax import masks
from flowjax.bijections.bijection import AbstractBijection
from flowjax.bijections.softplus import SoftPlus
from flowjax.bijections.tanh import LeakyTanh
from flowjax.bisection_search import AutoregressiveBisectionInverter
from flowjax.wrappers import BijectionReparam, WeightNormalization, Where


class _CallableToBijection(AbstractBijection):
    # Wrap a callable e.g. a function or a callable module. We assume the callable acts
    # on scalar values and log_det can be computed in a stable manner with jax.grad.

    fn: Callable
    shape: ClassVar[tuple] = ()
    cond_shape: ClassVar[None] = None

    def __init__(self, fn: Callable):
        if not callable(fn):
            raise TypeError(f"Expected callable, got {type(fn)}.")
        self.fn = fn

    def transform(self, x, condition=None):
        return self.fn(x)

    def transform_and_log_det(self, x, condition=None):
        y, grad = eqx.filter_value_and_grad(self.fn)(x)
        return y, jnp.log(jnp.abs(grad))

    def inverse(self, y, condition=None):
        raise NotImplementedError

    def inverse_and_log_det(self, y, condition=None):
        raise NotImplementedError


class BlockAutoregressiveNetwork(AbstractBijection):
    r"""Block Autoregressive Network (https://arxiv.org/abs/1904.04676).

    Note that in contrast to the original paper which uses tanh activations, by default
    we use :class:`~flowjax.bijections.tanh.LeakyTanh`. This ensures the codomain of the
    activation is the set of real values, which will ensure properly normalised
    densities (see https://github.com/danielward27/flowjax/issues/102).

    Args:
        key: Jax PRNGKey
        dim: Dimension of the distribution.
        cond_dim: Dimension of conditioning variables. Defaults to None.
        depth: Number of hidden layers in the network.
        block_dim: Block dimension (hidden layer size is `dim*block_dim`).
        activation: Activation function, either a scalar bijection or a callable that
            computes the activation for a scalar value. Note that the activation should
            be bijective to ensure invertibility of the network and in general should
            map real -> real to ensure that when transforming a distribution (either
            with the forward or inverse), the map is defined across the support of
            the base distribution. Defaults to ``LeakyTanh(3)``.
        inverter: Callable that implements the required numerical method to invert the
            ``BlockAutoregressiveNetwork`` bijection. Must have the signature
            ``inverter(bijection, y, condition=None)``. Defaults to
            ``AutoregressiveBisectionInverter``.
    """

    shape: tuple[int, ...]
    cond_shape: tuple[int, ...] | None
    depth: int
    layers: list
    cond_linear: eqx.nn.Linear | None
    block_dim: int
    activation: AbstractBijection
    inverter: Callable

    def __init__(
        self,
        key: PRNGKeyArray,
        *,
        dim: int,
        cond_dim: int | None = None,
        depth: int,
        block_dim: int,
        activation: AbstractBijection | Callable | None = None,
        inverter: Callable | None = None,
    ):
        key, subkey = jr.split(key)

        if activation is None:
            activation = LeakyTanh(3)
        elif isinstance(activation, AbstractBijection):
            if activation.shape != () or activation.cond_shape is not None:
                raise ValueError("Bijection must be unconditional with shape ().")
        else:
            activation = _CallableToBijection(activation)

        self.inverter = (
            AutoregressiveBisectionInverter() if inverter is None else inverter
        )

        layers_and_log_jac_fns = []
        if depth == 0:
            layers_and_log_jac_fns.append(
                block_autoregressive_linear(
                    key,
                    n_blocks=dim,
                    block_shape=(1, 1),
                ),
            )
        else:
            keys = random.split(key, depth + 1)

            block_shapes = [
                (block_dim, 1),
                *[(block_dim, block_dim)] * (depth - 1),
                (1, block_dim),
            ]

            for layer_key, block_shape in zip(keys, block_shapes, strict=True):
                layers_and_log_jac_fns.append(
                    block_autoregressive_linear(
                        layer_key,
                        n_blocks=dim,
                        block_shape=block_shape,
                    ),
                )

        if cond_dim is not None:
            layer0_out_dim = layers_and_log_jac_fns[0][0].out_features
            self.cond_linear = eqx.nn.Linear(
                cond_dim, layer0_out_dim, use_bias=False, key=subkey
            )
        else:
            self.cond_linear = None

        self.depth = depth
        self.layers = layers_and_log_jac_fns
        self.block_dim = block_dim
        self.shape = (dim,)
        self.cond_shape = None if cond_dim is None else (cond_dim,)
        self.activation = activation

    def transform(self, x, condition=None):
        for i, (layer, _) in enumerate(self.layers[:-1]):
            x = layer(x)
            if i == 0 and condition is not None:
                assert self.cond_linear is not None
                x += self.cond_linear(condition)
            x = eqx.filter_vmap(self.activation.transform)(x)
        return self.layers[-1][0](x)

    def transform_and_log_det(self, x, condition=None):
        log_dets_3ds = []
        for i, (linear, log_jacobian_fn) in enumerate(self.layers[:-1]):
            x = linear(x)
            if i == 0 and condition is not None:
                assert self.cond_linear is not None
                x += self.cond_linear(condition)
            log_dets_3ds.append(log_jacobian_fn(linear))
            x, log_det_3d = self._activation_and_log_jacobian_3d(x)
            log_dets_3ds.append(log_det_3d)

        linear, log_jacobian_fn = self.layers[-1]
        x = linear(x)
        log_dets_3ds.append(log_jacobian_fn(linear))

        log_det = log_dets_3ds[-1]
        for log_jacobian in reversed(log_dets_3ds[:-1]):
            log_det = logmatmulexp(log_det, log_jacobian)
        return x, log_det.sum()

    def inverse(self, y, condition=None):
        return self.inverter(self, y, condition)

    def inverse_and_log_det(self, y, condition=None):
        x = self.inverter(self, y, condition)
        _, forward_log_det = self.transform_and_log_det(x, condition)
        return x, -forward_log_det

    def _activation_and_log_jacobian_3d(self, x):
        """Compute activation and the log determinant (blocks, block_dim, block_dim)."""
        x, log_abs_grads = eqx.filter_vmap(self.activation.transform_and_log_det)(x)
        log_det_3d = jnp.full((self.shape[0], self.block_dim, self.block_dim), -jnp.inf)
        diag_idxs = jnp.arange(self.block_dim)
        log_det_3d = log_det_3d.at[:, diag_idxs, diag_idxs].set(
            log_abs_grads.reshape(self.shape[0], self.block_dim)
        )
        return x, log_det_3d


def block_autoregressive_linear(
    key: PRNGKeyArray,
    *,
    n_blocks: int,
    block_shape: tuple,
) -> tuple[eqx.nn.Linear, Callable]:
    """Block autoregressive linear layer (https://arxiv.org/abs/1904.04676).

    Returns:
        Tuple containing 1) an equinox.nn.Linear layer, with the weights wrapped
        in order to provide a block lower triangular mask and weight normalisation.
        2) callable taking the linear layer, returning the log block diagonal weights.

    Args:
        key: Random key.
        n_blocks: Number of diagonal blocks (dimension of original input).
        block_shape: The shape of the blocks.
    """
    out_features, in_features = (b * n_blocks for b in block_shape)
    linear = eqx.nn.Linear(in_features, out_features, key=key)
    block_diag_mask = masks.block_diag_mask(block_shape, n_blocks)
    block_tril_mask = masks.block_tril_mask(block_shape, n_blocks)

    weight = Where(block_tril_mask, linear.weight, 0)
    weight = Where(
        block_diag_mask,
        BijectionReparam(weight, SoftPlus(), invert_on_init=False),
        weight,
    )
    weight = WeightNormalization(weight)
    linear = eqx.tree_at(lambda linear: linear.weight, linear, replace=weight)

    def linear_to_log_block_diagonal(linear: eqx.nn.Linear):
        idxs = jnp.where(block_diag_mask, size=prod(block_shape) * n_blocks)
        jac_3d = linear.weight[idxs].reshape(n_blocks, *block_shape)
        return jnp.log(jac_3d)

    return linear, linear_to_log_block_diagonal


def logmatmulexp(x, y):
    """Numerically stable version of ``(x.log() @ y.log()).exp()``.

    From numpyro https://github.com/pyro-ppl/numpyro/blob/
    f2ff89a3a7147617e185eb51148eb15d56d44661/numpyro/distributions/util.py#L387.
    """
    x_shift = jax.lax.stop_gradient(jnp.amax(x, -1, keepdims=True))
    y_shift = jax.lax.stop_gradient(jnp.amax(y, -2, keepdims=True))
    xy = jnp.log(jnp.matmul(jnp.exp(x - x_shift), jnp.exp(y - y_shift)))
    return xy + x_shift + y_shift



"""Bijections that wrap jax function transforms (scan and vmap)."""

from collections.abc import Callable

import equinox as eqx
import jax.numpy as jnp
from jax.lax import scan
from jax.tree_util import tree_leaves, tree_map
from jaxtyping import PyTree

from flowjax import wrappers
from flowjax.bijections.bijection import AbstractBijection


class Scan(AbstractBijection):
    """Repeatedly apply the same bijection with different parameter values.

    Internally, uses `jax.lax.scan` to reduce compilation time. Often it is convenient
    to construct these using ``equinox.filter_vmap``.

    Args:
        bijection: A bijection, in which the arrays leaves have an additional leading
            axis to scan over. It is often can convenient to create compatible
            bijections with ``equinox.filter_vmap``.

    Example:
        Below is equivilent to ``Chain([Affine(p) for p in params])``.

        .. doctest::

            >>> from flowjax.bijections import Scan, Affine
            >>> import jax.numpy as jnp
            >>> import equinox as eqx
            >>> params = jnp.ones((3, 2))
            >>> affine = eqx.filter_vmap(Affine)(params)
            >>> affine = Scan(affine)
    """

    bijection: AbstractBijection

    def transform(self, x, condition=None):
        def step(x, bijection):
            return (bijection.transform(x, condition), None)

        y, _ = _filter_scan(step, x, self.bijection)
        return y

    def transform_and_log_det(self, x, condition=None):
        def step(carry, bijection):
            x, log_det = carry
            y, log_det_i = bijection.transform_and_log_det(x, condition)
            return ((y, log_det + log_det_i.sum()), None)

        (y, log_det), _ = _filter_scan(step, (x, 0), self.bijection)
        return y, log_det

    def inverse(self, y, condition=None):
        def step(y, bijection):
            return bijection.inverse(y, condition), None

        x, _ = _filter_scan(step, y, self.bijection, reverse=True)
        return x

    def inverse_and_log_det(self, y, condition=None):
        def step(carry, bijection):
            y, log_det = carry
            x, log_det_i = bijection.inverse_and_log_det(y, condition)
            return ((x, log_det + log_det_i.sum()), None)

        (y, log_det), _ = _filter_scan(step, (y, 0), self.bijection, reverse=True)
        return y, log_det

    @property
    def shape(self):
        return self.bijection.shape

    @property
    def cond_shape(self):
        return self.bijection.cond_shape


def _filter_scan(f, init, xs, *, reverse=False):
    params, static = eqx.partition(xs, filter_spec=eqx.is_array)

    def _scan_fn(carry, x):
        module = eqx.combine(x, static)
        carry, y = f(carry, module)
        return carry, y

    return scan(_scan_fn, init, params, reverse=reverse)


def _check_no_unwrappables(pytree):
    def _is_unwrappable(leaf):
        return isinstance(leaf, wrappers.AbstractUnwrappable)

    leaves = tree_leaves(pytree, is_leaf=_is_unwrappable)
    if any(_is_unwrappable(leaf) for leaf in leaves):
        raise ValueError(
            "In axes containing unwrappables is not supported. In axes must be "
            "specified to match the structure of the unwrapped pytree i.e after "
            "calling flowjax.wrappers.unwrap."
        )


class Vmap(AbstractBijection):
    """Applies vmap to bijection methods to add a batch dimension to the bijection.

    Args:
        bijection: The bijection to vectorize.
        in_axes: Specify which axes of the bijection parameters to vectorise over. It
            should be a PyTree of ``None``, ``int`` with the tree structure being a
            prefix of the bijection, or a callable mapping ``Leaf -> Union[None, int]``.
            Defaults to None. Note, if the bijection contains unwrappables, then in_axes
            should be specified for the unwrapped structure of the bijection.
        axis_size: The size of the new axis. This should be left unspecified if in_axes
            is provided, as the size can be inferred from the bijection parameters.
            Defaults to None.
        in_axes_condition: Optionally define an axis of the conditioning variable to
            vectorize over. Defaults to None.

    Example:
        The two most common use cases, are shown below:

        .. doctest::

            Add a batch dimension to a bijection, mapping over bijection parameters:

            >>> import jax.numpy as jnp
            >>> import equinox as eqx
            >>> from flowjax.bijections import Vmap, RationalQuadraticSpline, Affine
            >>> bijection = eqx.filter_vmap(
            ...    lambda: RationalQuadraticSpline(knots=5, interval=2),
            ...    axis_size=10
            ... )()
            >>> bijection = Vmap(bijection, in_axes=eqx.if_array(0))
            >>> bijection.shape
            (10,)

            Add a batch dimension to a bijection, broadcasting bijection parameters:
            >>> bijection = RationalQuadraticSpline(knots=5, interval=2)
            >>> bijection = Vmap(bijection, axis_size=10)
            >>> bijection.shape
            (10,)

        A more advanced use case is to create bijections with more fine grained control
        over parameter broadcasting. For example, the ``Affine`` constructor broadcasts
        the location and scale parameters during initialization. What if we want an
        ``Affine`` bijection, with a global scale parameter, but an elementwise location
        parameter? We could achieve this as follows.

            >>> from jax.tree_util import tree_map
            >>> from flowjax.wrappers import unwrap
            >>> bijection = Affine(jnp.zeros(()), jnp.ones(()))
            >>> bijection = eqx.tree_at(lambda bij: bij.loc, bijection, jnp.arange(3))
            >>> in_axes = tree_map(lambda _: None, unwrap(bijection))
            >>> in_axes = eqx.tree_at(
            ...     lambda bij: bij.loc, in_axes, 0, is_leaf=lambda x: x is None
            ...     )
            >>> bijection = Vmap(bijection, in_axes=in_axes)
            >>> bijection.shape
            (3,)
            >>> bijection.bijection.loc.shape
            (3,)
            >>> unwrap(bijection.bijection.scale).shape
            ()

            >>> x = jnp.ones(3)
            >>> bijection.transform(x)
            Array([1., 2., 3.], dtype=float32)

    """

    bijection: AbstractBijection
    in_axes: tuple
    axis_size: int
    cond_shape: tuple[int, ...] | None

    def __init__(
        self,
        bijection: AbstractBijection,
        *,
        in_axes: PyTree | None | int | Callable = None,
        axis_size: int | None = None,
        in_axes_condition: int | None = None,
    ):
        if in_axes is not None and axis_size is not None:
            raise ValueError("Cannot specify both in_axes and axis_size.")

        if axis_size is None:
            if in_axes is None:
                raise ValueError("Either axis_size or in_axes must be provided.")
            _check_no_unwrappables(in_axes)
            axis_size = _infer_axis_size_from_params(
                wrappers.unwrap(bijection), in_axes
            )

        self.in_axes = (in_axes, 0, in_axes_condition)
        self.bijection = bijection
        self.axis_size = axis_size
        self.cond_shape = self.get_cond_shape(in_axes_condition)

    def vmap(self, f: Callable):
        return eqx.filter_vmap(f, in_axes=self.in_axes, axis_size=self.axis_size)

    def transform(self, x, condition=None):
        def _transform(bijection, x, condition):
            return bijection.transform(x, condition)

        return self.vmap(_transform)(self.bijection, x, condition)

    def transform_and_log_det(self, x, condition=None):
        def _transform_and_log_det(bijection, x, condition):
            return bijection.transform_and_log_det(x, condition)

        y, log_det = self.vmap(_transform_and_log_det)(self.bijection, x, condition)
        return y, jnp.sum(log_det)

    def inverse(self, y, condition=None):
        def _inverse(bijection, x, condition):
            return bijection.inverse(x, condition)

        return self.vmap(_inverse)(self.bijection, y, condition)

    def inverse_and_log_det(self, y, condition=None):
        def _inverse_and_log_det(bijection, x, condition):
            return bijection.inverse_and_log_det(x, condition)

        x, log_det = self.vmap(_inverse_and_log_det)(self.bijection, y, condition)
        return x, jnp.sum(log_det)

    @property
    def shape(self):
        return (self.axis_size, *self.bijection.shape)

    def get_cond_shape(self, cond_ax):
        if self.bijection.cond_shape is None or cond_ax is None:
            return self.bijection.cond_shape
        return (
            *self.bijection.cond_shape[:cond_ax],
            self.axis_size,
            *self.bijection.cond_shape[cond_ax:],
        )


def _infer_axis_size_from_params(tree: PyTree, in_axes) -> int:
    axes = _resolve_vmapped_axes(tree, in_axes)
    axis_sizes = tree_leaves(
        tree_map(
            lambda leaf, ax: leaf.shape[ax] if ax is not None else None,
            tree,
            axes,
        ),
    )
    if len(axis_sizes) == 0:
        raise ValueError("in_axes did not map to any leaves to vectorize.")
    return axis_sizes[0]


def _resolve_vmapped_axes(pytree, in_axes):
    """Returns pytree with ints denoting vmapped dimensions."""

    # Adapted from equinox filter_vmap
    def _resolve_axis(in_axes, elem):
        if in_axes is None or isinstance(in_axes, int):
            return tree_map(lambda _: in_axes, elem)
        if callable(in_axes):
            return tree_map(in_axes, elem)
        raise TypeError("`in_axes` must consist of None, ints, and callables.")

    return tree_map(_resolve_axis, in_axes, pytree, is_leaf=lambda x: x is None)



"""Configuration file for the Sphinx documentation builder."""

import sys
import typing
from pathlib import Path

# tag used in jaxtyping prevent expanding cumbersome type aliases such as ArrayLike
typing.GENERATING_DOCUMENTATION = True


sys.path.insert(0, Path("..").resolve())

# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information

project = "FlowJax"
copyright = "2022, Daniel Ward"
author = "Daniel Ward"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "sphinx.ext.viewcode",
    "sphinx.ext.autodoc",
    "sphinx.ext.doctest",
    "sphinx.ext.intersphinx",
    "nbsphinx",
    "sphinx_copybutton",
    "sphinx.ext.napoleon",
    "sphinx_autodoc_typehints",
]

intersphinx_mapping = {
    "python": ("https://docs.python.org/3/", None),
    "jax": ("https://jax.readthedocs.io/en/latest/", None),
}

templates_path = ["_templates"]
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# napoleon_include_init_with_doc = False

# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

html_css_files = [
    "style.css",
]

html_theme_options = {
    "navigation_depth": 2,
}

pygments_style = "xcode"

copybutton_prompt_text = r">>> |\.\.\. "
copybutton_prompt_is_regexp = True

napolean_use_rtype = False
napoleon_attr_annotations = True

add_module_names = False



import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
import pytest
from jax.tree_util import tree_map

from flowjax.bijections import Affine, MaskedAutoregressive, Vmap
from flowjax.wrappers import unwrap


def test_vmap_uneven_init():
    "Tests adding a batch dimension to a particular leaf (parameter array)."
    bijection = Affine(jnp.zeros(()), jnp.ones(()))
    bijection = eqx.tree_at(lambda bij: bij.loc, bijection, jnp.arange(3))
    in_axes = tree_map(lambda _: None, unwrap(bijection))
    in_axes = eqx.tree_at(lambda bij: bij.loc, in_axes, 0, is_leaf=lambda x: x is None)
    bijection = Vmap(bijection, in_axes=in_axes)

    assert bijection.shape == (3,)
    assert bijection.bijection.loc.shape == (3,)
    assert unwrap(bijection.bijection.scale).shape == ()

    x = jnp.ones(3)
    expected = x + jnp.arange(3)
    assert bijection.transform(x) == pytest.approx(expected)


def test_vmap_error_with_unwrappable():
    bijection = Affine(jnp.zeros(1), jnp.ones(1))
    in_axes = tree_map(eqx.is_array, bijection)
    with pytest.raises(ValueError, match="unwrappable"):
        bijection = Vmap(bijection, in_axes=in_axes)


def test_vmap_condition_only():
    bijection = MaskedAutoregressive(
        jr.PRNGKey(0),
        transformer=Affine(),
        dim=3,
        cond_dim=4,
        nn_width=10,
        nn_depth=1,
    )

    with pytest.raises(
        ValueError,
        match="Either axis_size or in_axes must be provided.",
    ):
        bijection = Vmap(bijection, in_axes_condition=0)

    bijection = Vmap(bijection, in_axes_condition=1, axis_size=10)
    assert bijection.shape == (10, 3)
    assert bijection.cond_shape == (4, 10)

    x = jnp.ones((10, 3))
    condition = jnp.linspace(-2, 2, 10 * 4).reshape(4, 10)
    y = bijection.transform(x, condition)
    x_reconstructed = bijection.inverse(y, condition)
    assert x_reconstructed == pytest.approx(x)



"""Planar transform.

A layer in a planar flow introduced in https://arxiv.org/pdf/1505.05770.pdf.
"""

from collections.abc import Callable
from typing import ClassVar

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
from jax.nn import softplus
from jax.numpy.linalg import norm
from jaxtyping import Array, PRNGKeyArray

from flowjax.bijections.bijection import AbstractBijection


class Planar(AbstractBijection):
    r"""Planar bijection as used by https://arxiv.org/pdf/1505.05770.pdf.

    Uses the transformation :math:`y + u \cdot \text{tanh}(w \cdot x + b)`, where
    :math:`u \in \mathbb{R}^D, \ w \in \mathbb{R}^D` and :math:`b \in \mathbb{R}`. In
    the unconditional case, :math:`w`, :math:`u`  and :math:`b` are learned directly.
    In the conditional case they are parameterised by an MLP.

    Args:
        key: Jax random seed.
        dim: Dimension of the bijection.
        cond_dim: Dimension of extra conditioning variables. Defaults to None.
        **mlp_kwargs: Keyword arguments (excluding in_size and out_size) passed to
            the MLP (equinox.nn.MLP). Ignored when cond_dim is None.
    """

    shape: tuple[int, ...]
    cond_shape: tuple[int, ...] | None
    conditioner: Callable | None
    params: Array | None

    def __init__(
        self,
        key: PRNGKeyArray,
        *,
        dim: int,
        cond_dim: int | None = None,
        **mlp_kwargs,
    ):
        self.shape = (dim,)

        if cond_dim is None:
            self.params = 0.01 * jr.normal(key, (2 * dim + 1,))
            self.conditioner = None
            self.cond_shape = None
        else:
            self.params = None
            self.conditioner = eqx.nn.MLP(cond_dim, 2 * dim + 1, **mlp_kwargs, key=key)
            self.cond_shape = (cond_dim,)

    def transform(self, x, condition=None):
        return self.get_planar(condition).transform(x)

    def transform_and_log_det(self, x, condition=None):
        return self.get_planar(condition).transform_and_log_det(x)

    def inverse(self, y, condition=None):
        return self.get_planar(condition).inverse(y)

    def inverse_and_log_det(self, y, condition=None):
        return self.get_planar(condition).inverse_and_log_det(y)

    def get_planar(self, condition=None):
        """Get the planar bijection with the conditioning applied if conditional."""
        if self.cond_shape is not None:
            params = self.conditioner(condition)
        else:
            params = self.params
        dim = self.shape[0]
        assert params is not None
        w, u, bias = params[:dim], params[dim : 2 * dim], params[-1]
        return _UnconditionalPlanar(w, u, bias)


class _UnconditionalPlanar(AbstractBijection):
    """Unconditional planar bijection, used in Planar.

    Note act_scale (u in the paper) is unconstrained and the constraint to ensure
    invertiblitiy is applied in the ``get_act_scale``.
    """

    shape: tuple[int, ...]
    cond_shape: ClassVar[None] = None
    weight: Array
    _act_scale: Array
    bias: Array

    def __init__(self, weight, act_scale, bias):
        self.weight = weight
        self._act_scale = act_scale
        self.bias = bias
        self.shape = weight.shape

    def transform(self, x, condition=None):
        return x + self.get_act_scale() * jnp.tanh(self.weight @ x + self.bias)

    def transform_and_log_det(self, x, condition=None):
        u = self.get_act_scale()
        act = jnp.tanh(x @ self.weight + self.bias)
        y = x + u * act
        psi = (1 - act**2) * self.weight
        log_det = jnp.log(jnp.abs(1 + u @ psi))
        return y, log_det

    def get_act_scale(self):
        """Apply constraint to u to ensure invertibility.

        See appendix A1 in https://arxiv.org/pdf/1505.05770.pdf.
        """
        wtu = self._act_scale @ self.weight
        m_wtu = -1 + jnp.log(1 + softplus(wtu))
        return self._act_scale + (m_wtu - wtu) * self.weight / norm(self.weight) ** 2

    def inverse(self, y, condition=None):
        raise NotImplementedError(
            "The inverse planar transformation is not implemented.",
        )

    def inverse_and_log_det(self, y, condition=None):
        raise NotImplementedError(
            "The inverse planar transformation is not implemented.",
        )



""""Bisection search algorithm.

This is useful for inverting some bijections that do not have a known inverse.
"""

from collections.abc import Callable
from typing import NamedTuple

import equinox as eqx
import jax.numpy as jnp
from jax import lax
from jaxtyping import Array, Int, Real


class AutoregressiveBisectionInverter(eqx.Module):
    """Callable module to invert an autoregressive bijection using a bisection search.

    Note that if the inverse value is not within lower and upper, the bounds are
    dynamically adjusted using ``_adapt_interval_to_include_root``.

    Args:
        lower: Lower bound of the initial interval where the inverse value is expected.
        upper: Upper bound of the initial interval where the inverse value is expected.
        tol: Tolerance of solution. Note due to accumulation of errors at each step,
            the found solution may not fall within the given tolerance.
        max_iter: Maximum number of iterations to use.
    """

    lower: Real[Array, ""] = eqx.field(
        default_factory=lambda: -10.0, converter=jnp.asarray
    )
    upper: Real[Array, ""] = eqx.field(
        default_factory=lambda: 10.0, converter=jnp.asarray
    )
    tol: float = 1e-7
    max_iter: int = 200

    def __check_init__(self):
        if not self.lower < self.upper:
            raise ValueError("Lower must be less than upper.")
        if self.tol <= 0:
            raise ValueError("Tolerance must be positive.")
        if self.max_iter < 0:
            raise ValueError("max_iter must be a positive integer.")

    def __call__(self, bijection, y, condition=None):
        def fn(x):
            return bijection.transform(x, condition) - y

        return _autoregressive_bisection_search(
            autoregressive_fn=fn,
            lower=self.lower,
            upper=self.upper,
            tol=self.tol,
            length=bijection.shape[0],
            max_iter=self.max_iter,
        )


def _autoregressive_bisection_search(
    autoregressive_fn: Callable,
    *,
    lower: Real[Array, ""],
    upper: Real[Array, ""],
    tol: float,
    length: int,
    max_iter: int,
):
    """Bisection search for a monotonic increasing autoregressive function.

    We scan over the inputs finding the root element by element, assuming that
    each input only depends on previous inputs in the input array. This is useful
    for inverting some bijections without a known inverse, such as those used in block
    neural autoregressive flows. Note that tol refers to the tolerance used in each
    run of the bisection search. This means the found solution may not necessarily
    be within the tolerance for all elements, as errors can accumulate in each step.

    Args:
        autoregressive_fn: The monotonically increasing autoregressive function.
        length: The length of the vector input to the function.
        tol: Tolerance of a solution. Note due to accumulation of errors at each step,
            the found solution may not fall within the given tolerance.
        lower: Lower bound of the initial interval where the root is expected.
        upper: Upper bound of the initial interval where the root is expected.
        max_iter: Maximum number of iterations. Defaults to 200.
    """

    def scan_fn(init, _):
        y, i = init

        def scalar_fn(x):
            x = y.at[i].set(x)
            return autoregressive_fn(x)[i]

        root, *_ = _bisection_search(
            scalar_fn,
            tol=tol,
            lower=lower,
            upper=upper,
            max_iter=max_iter,
        )

        return (y.at[i].set(root), i + 1), None

    init = (jnp.full(length, (upper + lower) / 2), 0)
    (root, _), _ = lax.scan(f=scan_fn, init=init, xs=None, length=length)
    return root


def _bisection_search(
    func: Callable,
    *,
    lower: Real[Array, ""],
    upper: Real[Array, ""],
    tol: float,
    max_iter: int,
):
    """Bisection search algorithm to find a root of a scalar increasing function.

    Note that the implementation assumes that the function is monotonically increasing.
    If the initial limits do not include the root then the interval is adapted using
    ``expand_interval_to_include_root``.

    Args:
        func: Scalar increasing function to find the root for.
        lower: Lower bound of the initial interval where the root is expected.
        upper: Upper bound of the initial interval where the root is expected.
        tol: Tolerance level for the root. Defaults to 1e-6. Defaults to 1e-6.
        max_iter: Maximum number of iterations. Defaults to 200.

    Returns:
        A tuple ``(root, adapt_iterations, iterations)``. ``adapt_iterations`` will be
            zero if the root is beteween lower and upper).
    """
    if max_iter < 0:
        raise ValueError("max_iter must be positive.")
    if tol <= 0:
        raise ValueError("tol must be positive.")

    lower, upper, adapt_iterations = _adapt_interval_to_include_root(
        func,
        lower=lower,
        upper=upper,
    )

    def cond_fn(state):
        lower, upper, iterations = state
        return jnp.logical_and((upper - lower) > 2 * tol, iterations < max_iter)

    def body_fn(state):
        lower, upper, iterations = state
        midpoint = (lower + upper) / 2
        sign = jnp.sign(func(midpoint))
        lower = jnp.where(sign == 1, lower, midpoint)
        upper = jnp.where(sign == 1, midpoint, upper)

        # In case we hit the root exactly
        lower = jnp.where(sign == 0, midpoint, lower)
        upper = jnp.where(sign == 0, midpoint, upper)
        return lower, upper, iterations + 1

    init_state = (lower, upper, 0)
    lower, upper, iterations = lax.while_loop(cond_fn, body_fn, init_state)
    root = (lower + upper) / 2
    return root, adapt_iterations, iterations


def _adapt_interval_to_include_root(
    func,
    *,
    lower: Real[Array, ""],
    upper: Real[Array, ""],
    expand_factor: float = 2.0,
):
    """Dyamically adjust the interval to include the root of an increasing function.

    Note we do not currently perform any argument checking as it is challenging to
    perform checks that rely on array values with jax. It is the users responsibility
    to ensure lower is less than upper, and the function is increasing.

    Args:
        func: A scalar increasing function.
        lower: Lower component of interval. Must be less than upper.
        upper: Upper component of interval. Must be greater than upper.
        expand_factor: How much to (multiplicatively) increase the adjustment by on
            each iteration. The magnitude of the adjustment of the necessary bound is
            given by ``(init_upper - init_lower)*expand_factor**iteration``. Defaults to
            2.0.
    """
    fn_lower, fn_upper = func(lower), func(upper)

    class _State(NamedTuple):
        lower: Real[Array, ""]
        upper: Real[Array, ""]
        expand_by: Real[Array, ""]
        lower_fn_sign: Real[Array, ""]
        upper_fn_sign: Real[Array, ""]
        iteration: Int[Array, ""] = jnp.array(0)

    def cond_fn(state):
        return state.lower_fn_sign == state.upper_fn_sign

    def body_fn(state):
        sign = state.lower_fn_sign  # Note we know the signs match from cond_fn
        lower_update = jnp.where(sign == 1, state.lower - state.expand_by, state.upper)
        upper_update = jnp.where(sign == 1, state.lower, state.upper + state.expand_by)
        return _State(
            lower=lower_update,
            upper=upper_update,
            expand_by=state.expand_by * expand_factor,
            lower_fn_sign=jnp.sign(func(lower_update)),
            upper_fn_sign=jnp.sign(func(upper_update)),
            iteration=state.iteration + 1,
        )

    lower, upper = jnp.asarray(lower, float), jnp.asarray(upper, float)

    init_state = _State(
        lower=lower,
        upper=upper,
        expand_by=upper - lower,
        lower_fn_sign=jnp.sign(fn_lower),
        upper_fn_sign=jnp.sign(fn_upper),
    )
    state = lax.while_loop(cond_fn, body_fn, init_state)

    # In case we land on root exactly, return lower==upper==root
    lower, upper = state.lower, state.upper
    lower = jnp.where(state.upper_fn_sign == 0, upper, lower)
    upper = jnp.where(state.lower_fn_sign == 0, lower, upper)
    return lower, upper, state.iteration



import jax.numpy as jnp
from jax import random

from flowjax.bijections.masked_autoregressive import masked_autoregressive_mlp
from flowjax.masks import rank_based_mask
from flowjax.wrappers import unwrap


def test_masked_autoregressive_mlp():
    key = random.PRNGKey(0)
    in_size = 4
    in_ranks = jnp.arange(in_size)
    hidden_ranks = jnp.arange(6) % in_size
    out_ranks = jnp.arange(in_size).repeat(2)

    # Extract masks before unwrapping
    mlp = masked_autoregressive_mlp(in_ranks, hidden_ranks, out_ranks, depth=3, key=key)
    masks = [layer.weight.cond for layer in mlp.layers]

    mlp = unwrap(mlp)
    x = jnp.ones(in_size)
    y = mlp(x)
    assert y.shape == out_ranks.shape

    # Check connectivity matrix matches expected dependency structure
    a = masks[-1].astype(int)
    for mask in reversed(masks[:-1]):
        a = a @ mask.astype(int)

    expected = rank_based_mask(in_ranks, out_ranks)
    assert jnp.all((a > 0) == expected)



import jax
import jax.numpy as jnp
import pytest

from flowjax.bijections import Affine, Partial, Permute

test_cases = {
    # name: idx, expected
    "int": (1, jnp.array([False, True, False, False])),
    "bool_array": (
        jnp.array([True, False, True, False]),
        jnp.array([True, False, True, False]),
    ),
    "int_arry": (jnp.array([0, 2]), jnp.array([True, False, True, False])),
}


@pytest.mark.parametrize(
    ("idx", "expected"),
    test_cases.values(),
    ids=test_cases.keys(),
)
def test_partial(idx, expected):
    "Check values only change where we expect."
    x = jnp.zeros(4)
    shape = x[idx].shape
    bijection = Partial(Affine(jnp.ones(shape)), idx, x.shape)
    y = bijection.transform(x)
    assert jnp.all((x != y) == expected)


def test_Permute_argcheck():
    with pytest.raises(jax.lib.xla_extension.XlaRuntimeError):
        Permute(jnp.array([0, 0]))



"""Implemenetation of Coupling flow layer with arbitrary transformer.

Ref: https://arxiv.org/abs/1605.08803.
"""

from collections.abc import Callable

import equinox as eqx
import jax.nn as jnn
import jax.numpy as jnp
from jaxtyping import PRNGKeyArray

from flowjax.bijections.bijection import AbstractBijection
from flowjax.bijections.jax_transforms import Vmap
from flowjax.utils import Array, get_ravelled_pytree_constructor


class Coupling(AbstractBijection):
    """Coupling layer implementation (https://arxiv.org/abs/1605.08803).

    Args:
        key: Jax PRNGKey
        transformer: Unconditional bijection with shape () to be parameterised by the
            conditioner neural netork. Parameters wrapped with ``NonTrainable``
            are excluded from being parameterized.
        untransformed_dim: Number of untransformed conditioning variables (e.g. dim//2).
        dim: Total dimension.
        cond_dim: Dimension of additional conditioning variables. Defaults to None.
        nn_width: Neural network hidden layer width.
        nn_depth: Neural network hidden layer size.
        nn_activation: Neural network activation function. Defaults to jnn.relu.
    """

    shape: tuple[int, ...]
    cond_shape: tuple[int, ...] | None
    untransformed_dim: int
    dim: int
    transformer_constructor: Callable
    conditioner: eqx.nn.MLP

    def __init__(
        self,
        key: PRNGKeyArray,
        *,
        transformer: AbstractBijection,
        untransformed_dim: int,
        dim: int,
        cond_dim: int | None = None,
        nn_width: int,
        nn_depth: int,
        nn_activation: Callable = jnn.relu,
    ):
        if transformer.shape != () or transformer.cond_shape is not None:
            raise ValueError(
                "Only unconditional transformers with shape () are supported.",
            )

        constructor, num_params = get_ravelled_pytree_constructor(transformer)

        self.transformer_constructor = constructor
        self.untransformed_dim = untransformed_dim
        self.dim = dim
        self.shape = (dim,)
        self.cond_shape = (cond_dim,) if cond_dim is not None else None

        conditioner_output_size = num_params * (dim - untransformed_dim)

        self.conditioner = eqx.nn.MLP(
            in_size=(
                untransformed_dim if cond_dim is None else untransformed_dim + cond_dim
            ),
            out_size=conditioner_output_size,
            width_size=nn_width,
            depth=nn_depth,
            activation=nn_activation,
            key=key,
        )

    def transform(self, x, condition=None):
        x_cond, x_trans = x[: self.untransformed_dim], x[self.untransformed_dim :]
        nn_input = x_cond if condition is None else jnp.hstack((x_cond, condition))
        transformer_params = self.conditioner(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        y_trans = transformer.transform(x_trans)
        return jnp.hstack((x_cond, y_trans))

    def transform_and_log_det(self, x, condition=None):
        x_cond, x_trans = x[: self.untransformed_dim], x[self.untransformed_dim :]
        nn_input = x_cond if condition is None else jnp.hstack((x_cond, condition))
        transformer_params = self.conditioner(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        y_trans, log_det = transformer.transform_and_log_det(x_trans)
        y = jnp.hstack((x_cond, y_trans))
        return y, log_det

    def inverse(self, y, condition=None):
        x_cond, y_trans = y[: self.untransformed_dim], y[self.untransformed_dim :]
        nn_input = x_cond if condition is None else jnp.concatenate((x_cond, condition))
        transformer_params = self.conditioner(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        x_trans = transformer.inverse(y_trans)
        return jnp.hstack((x_cond, x_trans))

    def inverse_and_log_det(self, y, condition=None):
        x_cond, y_trans = y[: self.untransformed_dim], y[self.untransformed_dim :]
        nn_input = x_cond if condition is None else jnp.concatenate((x_cond, condition))
        transformer_params = self.conditioner(nn_input)
        transformer = self._flat_params_to_transformer(transformer_params)
        x_trans, log_det = transformer.inverse_and_log_det(y_trans)
        x = jnp.hstack((x_cond, x_trans))
        return x, log_det

    def _flat_params_to_transformer(self, params: Array):
        """Reshape to dim X params_per_dim, then vmap."""
        dim = self.dim - self.untransformed_dim
        transformer_params = jnp.reshape(params, (dim, -1))
        transformer = eqx.filter_vmap(self.transformer_constructor)(transformer_params)
        return Vmap(transformer, in_axes=eqx.if_array(0))



"""Premade versions of common flow architetctures from ``flowjax.flows``.

All these functions return a :class:`~flowjax.distributions.Transformed` distribution.
"""

# Note that here although we could chain arbitrary bijections using `Chain`, here,
# we generally opt to use `Scan`, which avoids excessive compilation
# when the flow layers share the same structure.
from collections.abc import Callable
from functools import partial

import equinox as eqx
import jax.nn as jnn
import jax.numpy as jnp
import jax.random as jr
from equinox.nn import Linear
from jax.nn.initializers import glorot_uniform
from jaxtyping import PRNGKeyArray

from flowjax.bijections import (
    AbstractBijection,
    AdditiveCondition,
    Affine,
    BlockAutoregressiveNetwork,
    Chain,
    Coupling,
    Flip,
    Invert,
    LeakyTanh,
    Loc,
    MaskedAutoregressive,
    Permute,
    Planar,
    RationalQuadraticSpline,
    Scan,
    SoftPlus,
    TriangularAffine,
    Vmap,
)
from flowjax.distributions import AbstractDistribution, Transformed
from flowjax.wrappers import BijectionReparam, NonTrainable, WeightNormalization


def _affine_with_min_scale(min_scale: float = 1e-2) -> Affine:
    scale_reparam = Chain([SoftPlus(), NonTrainable(Loc(min_scale))])
    return eqx.tree_at(
        where=lambda aff: aff.scale,
        pytree=Affine(),
        replace=BijectionReparam(jnp.array(1), scale_reparam),
    )


def coupling_flow(
    key: PRNGKeyArray,
    *,
    base_dist: AbstractDistribution,
    transformer: AbstractBijection | None = None,
    cond_dim: int | None = None,
    flow_layers: int = 8,
    nn_width: int = 50,
    nn_depth: int = 1,
    nn_activation: Callable = jnn.relu,
    invert: bool = True,
) -> Transformed:
    """Create a coupling flow (https://arxiv.org/abs/1605.08803).

    Args:
        key: Jax random number generator key.
        base_dist: Base distribution, with ``base_dist.ndim==1``.
        transformer: Bijection to be parameterised by conditioner. Defaults to
            affine.
        cond_dim: Dimension of conditioning variables. Defaults to None.
        flow_layers: Number of coupling layers. Defaults to 8.
        nn_width: Conditioner hidden layer size. Defaults to 50.
        nn_depth: Conditioner depth. Defaults to 1.
        nn_activation: Conditioner activation function. Defaults to jnn.relu.
        invert: Whether to invert the bijection. Broadly, True will prioritise a faster
            `inverse` methods, leading to faster `log_prob`, False will prioritise
            faster `transform` methods, leading to faster `sample`. Defaults to True.
    """
    if transformer is None:
        transformer = _affine_with_min_scale()

    dim = base_dist.shape[-1]

    def make_layer(key):  # coupling layer + permutation
        bij_key, perm_key = jr.split(key)
        bijection = Coupling(
            key=bij_key,
            transformer=transformer,
            untransformed_dim=dim // 2,
            dim=dim,
            cond_dim=cond_dim,
            nn_width=nn_width,
            nn_depth=nn_depth,
            nn_activation=nn_activation,
        )
        return _add_default_permute(bijection, dim, perm_key)

    keys = jr.split(key, flow_layers)
    layers = eqx.filter_vmap(make_layer)(keys)
    bijection = Invert(Scan(layers)) if invert else Scan(layers)
    return Transformed(base_dist, bijection)


def masked_autoregressive_flow(
    key: PRNGKeyArray,
    *,
    base_dist: AbstractDistribution,
    transformer: AbstractBijection | None = None,
    cond_dim: int | None = None,
    flow_layers: int = 8,
    nn_width: int = 50,
    nn_depth: int = 1,
    nn_activation: Callable = jnn.relu,
    invert: bool = True,
) -> Transformed:
    """Masked autoregressive flow.

    Parameterises a transformer bijection with an autoregressive neural network.
    Refs: https://arxiv.org/abs/1606.04934; https://arxiv.org/abs/1705.07057v4.

    Args:
        key: Random seed.
        base_dist: Base distribution, with ``base_dist.ndim==1``.
        transformer: Bijection parameterised by autoregressive network. Defaults to
            affine.
        cond_dim: Dimension of the conditioning variable. Defaults to None.
        flow_layers: Number of flow layers. Defaults to 8.
        nn_width: Number of hidden layers in neural network. Defaults to 50.
        nn_depth: Depth of neural network. Defaults to 1.
        nn_activation: _description_. Defaults to jnn.relu.
        invert: Whether to invert the bijection. Broadly, True will prioritise a faster
            inverse, leading to faster `log_prob`, False will prioritise faster forward,
            leading to faster `sample`. Defaults to True.
    """
    if transformer is None:
        transformer = _affine_with_min_scale()

    dim = base_dist.shape[-1]

    def make_layer(key):  # masked autoregressive layer + permutation
        bij_key, perm_key = jr.split(key)
        bijection = MaskedAutoregressive(
            key=bij_key,
            transformer=transformer,
            dim=dim,
            cond_dim=cond_dim,
            nn_width=nn_width,
            nn_depth=nn_depth,
            nn_activation=nn_activation,
        )
        return _add_default_permute(bijection, dim, perm_key)

    keys = jr.split(key, flow_layers)
    layers = eqx.filter_vmap(make_layer)(keys)
    bijection = Invert(Scan(layers)) if invert else Scan(layers)
    return Transformed(base_dist, bijection)


def block_neural_autoregressive_flow(
    key: PRNGKeyArray,
    *,
    base_dist: AbstractDistribution,
    cond_dim: int | None = None,
    nn_depth: int = 1,
    nn_block_dim: int = 8,
    flow_layers: int = 1,
    invert: bool = True,
    activation: AbstractBijection | Callable | None = None,
    inverter: Callable | None = None,
) -> Transformed:
    """Block neural autoregressive flow (BNAF) (https://arxiv.org/abs/1904.04676).

    Each flow layer contains a
    :class:`~flowjax.bijections.block_autoregressive_network.BlockAutoregressiveNetwork`
    bijection. The bijection does not have an analytic inverse, so must be inverted
    using numerical methods (by default a bisection search). Note that this means
    that only one of ``log_prob`` or ``sample{_and_log_prob}`` can be efficient,
    controlled by the ``invert`` argument.

    Args:
        key: Jax PRNGKey.
        base_dist: Base distribution, with ``base_dist.ndim==1``.
        cond_dim: Dimension of conditional variables. Defaults to None.
        nn_depth: Number of hidden layers within the networks. Defaults to 1.
        nn_block_dim: Block size. Hidden layer width is dim*nn_block_dim. Defaults to 8.
        flow_layers: Number of BNAF layers. Defaults to 1.
        invert: Use `True` for efficient ``log_prob`` (e.g. when fitting by maximum
            likelihood), and `False` for efficient ``sample`` and
            ``sample_and_log_prob`` methods (e.g. for fitting variationally).
        activation: Activation function used within block neural autoregressive
            networks. Note this should be bijective and in some use cases should map
            real -> real. For more information, see
            :class:`~flowjax.bijections.block_autoregressive_network.BlockAutoregressiveNetwork`.
            Defaults to :class:`~flowjax.bijections.tanh.LeakyTanh`.
        inverter: Callable that implements the required numerical method to invert the
            ``BlockAutoregressiveNetwork`` bijection. Must have the signature
            ``inverter(bijection, y, condition=None)``. Defaults to using a bisection
            search via ``AutoregressiveBisectionInverter``.
    """

    def make_layer(key):  # bnaf layer + permutation
        bij_key, perm_key = jr.split(key)
        bijection = BlockAutoregressiveNetwork(
            bij_key,
            dim=base_dist.shape[-1],
            cond_dim=cond_dim,
            depth=nn_depth,
            block_dim=nn_block_dim,
            activation=activation,
            inverter=inverter,
        )
        return _add_default_permute(bijection, base_dist.shape[-1], perm_key)

    keys = jr.split(key, flow_layers)
    layers = eqx.filter_vmap(make_layer)(keys)
    bijection = Invert(Scan(layers)) if invert else Scan(layers)
    return Transformed(base_dist, bijection)


def planar_flow(
    key: PRNGKeyArray,
    *,
    base_dist: AbstractDistribution,
    cond_dim: int | None = None,
    flow_layers: int = 8,
    invert: bool = True,
    **mlp_kwargs,
) -> Transformed:
    """Planar flow as introduced in https://arxiv.org/pdf/1505.05770.pdf.

    This alternates between :class:`~flowjax.bijections.planar.Planar` layers and
    permutations. Note the definition here is inverted compared to the original paper.

    Args:
        key: Jax PRNGKey.
        base_dist: Base distribution, with ``base_dist.ndim==1``.
        cond_dim: Dimension of conditioning variables. Defaults to None.
        flow_layers: Number of flow layers. Defaults to 8.
        invert: Whether to invert the bijection. Broadly, True will prioritise a faster
            `inverse` methods, leading to faster `log_prob`, False will prioritise
            faster `transform` methods, leading to faster `sample`. Defaults to True.
        **mlp_kwargs: Keyword arguments (excluding in_size and out_size) passed to
            the MLP (equinox.nn.MLP). Ignored when cond_dim is None.
    """

    def make_layer(key):  # Planar layer + permutation
        bij_key, perm_key = jr.split(key)
        bijection = Planar(
            bij_key,
            dim=base_dist.shape[-1],
            cond_dim=cond_dim,
            **mlp_kwargs,
        )
        return _add_default_permute(bijection, base_dist.shape[-1], perm_key)

    keys = jr.split(key, flow_layers)
    layers = eqx.filter_vmap(make_layer)(keys)
    bijection = Invert(Scan(layers)) if invert else Scan(layers)
    return Transformed(base_dist, bijection)


def triangular_spline_flow(
    key: PRNGKeyArray,
    *,
    base_dist: AbstractDistribution,
    cond_dim: int | None = None,
    flow_layers: int = 8,
    knots: int = 8,
    tanh_max_val: float | int = 3.0,
    invert: bool = True,
    init: Callable | None = None,
) -> Transformed:
    """Triangular spline flow.

    A single layer consists where each layer consists of a triangular affine
    transformation with weight normalisation, and an elementwise rational quadratic
    spline. Tanh is used to constrain to the input to [-1, 1] before spline
    transformations.

    Args:
        key: Jax random seed.
        base_dist: Base distribution, with ``base_dist.ndim==1``.
        cond_dim: The number of conditioning features. Defaults to None.
        flow_layers: Number of flow layers. Defaults to 8.
        knots: Number of knots in the splines. Defaults to 8.
        tanh_max_val: Maximum absolute value beyond which we use linear "tails" in the
            tanh function. Defaults to 3.0.
        invert: Whether to invert the bijection before transforming the base
            distribution. Defaults to True.
        init: Initialisation method for the lower triangular weights.
            Defaults to glorot_uniform().
    """
    init = init if init is not None else glorot_uniform()
    dim = base_dist.shape[-1]

    def get_splines():
        fn = partial(RationalQuadraticSpline, knots=knots, interval=1)
        spline = eqx.filter_vmap(fn, axis_size=dim)()
        return Vmap(spline, in_axes=eqx.if_array(0))

    def make_layer(key):
        lt_key, perm_key, cond_key = jr.split(key, 3)
        weights = init(lt_key, (dim, dim))
        lt_weights = weights.at[jnp.diag_indices(dim)].set(1)
        tri_aff = TriangularAffine(jnp.zeros(dim), lt_weights)
        tri_aff = eqx.tree_at(
            lambda t: t.triangular, tri_aff, replace_fn=WeightNormalization
        )
        bijections = [
            LeakyTanh(tanh_max_val, (dim,)),
            get_splines(),
            Invert(LeakyTanh(tanh_max_val, (dim,))),
            tri_aff,
        ]

        if cond_dim is not None:
            linear_condition = AdditiveCondition(
                Linear(cond_dim, dim, use_bias=False, key=cond_key),
                (dim,),
                (cond_dim,),
            )
            bijections.append(linear_condition)

        bijection = Chain(bijections)
        return _add_default_permute(bijection, dim, perm_key)

    keys = jr.split(key, flow_layers)
    layers = eqx.filter_vmap(make_layer)(keys)
    bijection = Invert(Scan(layers)) if invert else Scan(layers)
    return Transformed(base_dist, bijection)


def _add_default_permute(bijection: AbstractBijection, dim: int, key: PRNGKeyArray):
    if dim == 1:
        return bijection
    if dim == 2:
        return Chain([bijection, Flip((dim,))]).merge_chains()

    perm = Permute(jr.permutation(key, jnp.arange(dim)))
    return Chain([bijection, perm]).merge_chains()



from collections.abc import Callable
from math import prod
from typing import NamedTuple

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
import numpy as np
import pytest
from jax.scipy.stats import multivariate_normal

from flowjax.bijections import Exp
from flowjax.distributions import (
    AbstractDistribution,
    AbstractTransformed,
    Cauchy,
    Exponential,
    Gumbel,
    Laplace,
    Logistic,
    LogNormal,
    MultivariateNormal,
    Normal,
    StandardNormal,
    StudentT,
    Transformed,
    Uniform,
    VmapMixture,
    _StandardCauchy,
    _StandardGumbel,
    _StandardLaplace,
    _StandardLogistic,
    _StandardStudentT,
    _StandardUniform,
)

# This sets up a number of constructors shape -> instance for testing
# the generic API of Distribution classes.
# Note we do not test the private "Standard" distributions, assuming they are
# sufficiently tested by their loc, scale public counterparts.
_test_distributions = {
    # flowjax.distributions
    "StandardNormal": StandardNormal,
    "Normal": lambda shape: Normal(jnp.zeros(shape)),
    "_StandardUniform": _StandardUniform,
    "Uniform": lambda shape: Uniform(jnp.zeros(shape), 1),
    "_StandardGumbel": _StandardGumbel,
    "Gumbel": lambda shape: Gumbel(jnp.zeros(shape)),
    "_StandardCauchy": _StandardCauchy,
    "Cauchy": lambda shape: Cauchy(jnp.zeros(shape)),
    "_StandardStudentT": lambda shape: _StandardStudentT(jnp.ones(shape)),
    "StudentT": lambda shape: StudentT(jnp.ones(shape)),
    "LogNormal": lambda shape: LogNormal(jnp.ones(shape), 2),
    "_StandardLaplace": _StandardLaplace,
    "Laplace": lambda shape: Laplace(jnp.ones(shape)),
    "Exponential": lambda shape: Exponential(jnp.ones(shape)),
    "_StandardLogistic": _StandardLogistic,
    "Logistic": lambda shape: Logistic(jnp.ones(shape)),
    "VmapMixture": lambda shape: VmapMixture(
        eqx.filter_vmap(Normal)(jnp.arange(3 * prod(shape)).reshape(3, *shape)),
        weights=jnp.arange(3) + 1,
    ),
}


_test_distributions = [pytest.param(v, id=k) for k, v in _test_distributions.items()]
_test_shapes = [(), (2,), (2, 3)]


@pytest.mark.parametrize("distribution", _test_distributions)
@pytest.mark.parametrize("shape", _test_shapes)
def test_sample(distribution, shape):
    d = distribution(shape=shape)
    sample = d.sample(jr.PRNGKey(0))
    assert sample.shape == shape

    sample_shape = (1, 2)
    sample = d.sample(jr.PRNGKey(0), sample_shape)
    assert sample.shape == sample_shape + shape


@pytest.mark.parametrize("distribution", _test_distributions)
@pytest.mark.parametrize("shape", _test_shapes)
def test_log_prob(distribution, shape):
    d = distribution(shape=shape)
    x = d.sample(jr.PRNGKey(0))

    assert d.log_prob(x).shape == ()

    sample_shape = (1, 2)
    x = d.sample(jr.PRNGKey(0), sample_shape)
    assert d.log_prob(x).shape == sample_shape

    # test arraylike input
    assert jnp.all(d.log_prob(np.array(x)) == d.log_prob(x))

    if d.shape == ():
        assert d.log_prob(jnp.array(0)) == d.log_prob(0)


def test_uniform_params():
    dist = Uniform(
        jnp.array([1.0, 2.0]),
        jnp.array([3.0, 4.0]),
    )

    assert dist.minval == pytest.approx(jnp.array([1.0, 2.0]))
    assert dist.maxval == pytest.approx(jnp.array([3.0, 4.0]))


# Since the broadcasting behaviour is shared by all distributions
# we test it for a single unconditional and conditional distribution only.

dist_shape, sample_shape, condition_shape = ([(), (3, 4)] for _ in range(3))


class _TestDist(AbstractDistribution):
    "Toy distribution object, for testing of distribution broadcasting."
    shape: tuple[int, ...]
    cond_shape: tuple[int, ...] | None

    def __init__(self, shape, cond_shape=None):
        self.shape = shape
        self.cond_shape = cond_shape

    def _log_prob(self, x, condition=None):
        return jnp.zeros(())

    def _sample(self, key, condition=None):
        return jnp.zeros(self.shape)

    def _sample_and_log_prob(self, key, condition=None):
        return jnp.zeros(self.shape), jnp.zeros(())


def test_multivariate_normal():
    loc = jnp.arange(2)
    cov = jnp.array([[2, 0.5], [0.5, 3]])
    mvn = MultivariateNormal(loc, cov)

    key = jr.PRNGKey(0)
    sample = mvn.sample(key)
    expected = pytest.approx(multivariate_normal.logpdf(sample, loc, cov))
    assert mvn.log_prob(sample) == expected
    assert mvn.covariance == pytest.approx(cov)


@pytest.mark.parametrize("dist_shape", dist_shape)
@pytest.mark.parametrize("sample_shape", sample_shape)
def test_broadcasting_unconditional(dist_shape, sample_shape):
    d = _TestDist(dist_shape)
    samples = d.sample(jr.PRNGKey(0), sample_shape)
    assert samples.shape == sample_shape + dist_shape

    log_probs = d.log_prob(samples)
    assert log_probs.shape == sample_shape


@pytest.mark.parametrize("dist_shape", dist_shape)
@pytest.mark.parametrize("sample_shape", sample_shape)
@pytest.mark.parametrize("condition_shape", condition_shape)
@pytest.mark.parametrize(
    "leading_cond_shape",
    [(), (3, 4)],
)  # Additional leading dimensions in condition
def test_broadcasting_conditional(
    dist_shape,
    sample_shape,
    condition_shape,
    leading_cond_shape,
):
    key = jr.PRNGKey(0)
    d = _TestDist(dist_shape, condition_shape)
    condition = jnp.zeros(leading_cond_shape + condition_shape)
    samples = d.sample(key, sample_shape, condition)
    assert samples.shape == sample_shape + leading_cond_shape + dist_shape

    log_probs = d.log_prob(samples, condition)
    assert log_probs.shape == sample_shape + leading_cond_shape

    samples, log_probs = d.sample_and_log_prob(key, sample_shape, condition)
    assert samples.shape == sample_shape + leading_cond_shape + dist_shape
    assert log_probs.shape == sample_shape + leading_cond_shape


test_cases = [
    StandardNormal(
        (2, 2),
    ),  # Won't have custom sample_and_log_prob implementation as not Transformed
    Normal(jnp.ones((2, 2))),  # Will have custom implementation as is Transformed
]


@pytest.mark.parametrize("dist", test_cases)
def test_sample_and_log_prob(dist):
    # We test broadcasting behaviour seperately above.
    # Just check consistency to seperately using methods
    key = jr.PRNGKey(0)
    x_naive = dist._sample(key)
    lp_naive = dist._log_prob(x_naive)
    x, lp = dist._sample_and_log_prob(key)
    assert x == pytest.approx(x_naive)
    assert lp == pytest.approx(lp_naive)


def test_transformed_merge_transforms():
    shape = (3, 3)
    nested = Transformed(Normal(jnp.ones(shape)), Exp(shape))
    unnested = nested.merge_transforms()

    assert isinstance(nested.base_dist, AbstractTransformed)
    assert not isinstance(unnested.base_dist, AbstractTransformed)

    key = jr.PRNGKey(0)
    sample = unnested.sample(key)
    assert pytest.approx(sample) == nested.sample(key)

    assert pytest.approx(nested.log_prob(sample)) == unnested.log_prob(sample)


class _TestCase(NamedTuple):
    name: str
    method: Callable
    args: tuple
    error: type
    match: str


test_cases = [
    _TestCase(
        name="missing condition sample",
        method=_TestDist((), ()).sample,
        args=(jr.PRNGKey(0),),
        error=TypeError,
        match="Expected condition to be arraylike",
    ),
    _TestCase(
        name="wrong x shape log prob",
        method=_TestDist((2,), None).log_prob,
        args=(jnp.ones((3, 3)), None),
        error=ValueError,
        match="Expected trailing dimensions matching",
    ),
    _TestCase(
        name="wrong condition shape log prob",
        method=_TestDist((), (2, 3)).log_prob,
        args=(0, jnp.ones((3, 3))),
        error=ValueError,
        match="Expected trailing dimensions matching",
    ),
]


@pytest.mark.parametrize("test_case", test_cases, ids=[t.name for t in test_cases])
def test_method_errors(test_case):
    with pytest.raises(test_case.error, match=test_case.match):
        test_case.method(*test_case.args)


@pytest.mark.parametrize("weights", [jnp.array([0.25, 0.75]), jnp.array([1.0, 3])])
def test_vmap_mixture(weights):
    x = 3
    gaussian_mixture = VmapMixture(
        eqx.filter_vmap(Normal)(jnp.arange(2)),
        weights=weights,
    )
    normalized_weights = weights / weights.sum()
    expected = jnp.log(
        normalized_weights[0] * jnp.exp(Normal(0).log_prob(x))
        + normalized_weights[1] * jnp.exp(Normal(1).log_prob(x))
    )
    assert pytest.approx(expected) == gaussian_mixture.log_prob(x)



import jax.numpy as jnp
import pytest

from flowjax.utils import _get_ufunc_signature, merge_cond_shapes

test_cases = [
    # arrays, expected_shape
    ((jnp.ones(3), jnp.ones(3)), (3,)),
    ((jnp.ones(3), 1.0), (3,)),
    ((1.0, jnp.ones(3)), (3,)),
    ((1.0, 1.0), (1,)),
    (((1.0),), (1,)),
]

test_cases = [
    # input, expected
    ([(2, 3), (2, 3), (2, 3)], (2, 3)),
    ([(), ()], ()),
    ([None, (2, 3), (2, 3)], (2, 3)),
    ([(7,), (7,)], (7,)),
]


@pytest.mark.parametrize(("input_", "expected"), test_cases)
def test_merge_shapes(input_, expected):
    assert merge_cond_shapes(input_) == expected


test_cases_error = [[(2, 3), (2, 1)], [(2, 3), (4, 2, 3)]]


@pytest.mark.parametrize("input_", test_cases_error)
def test_merge_shapes_errors(input_):
    with pytest.raises(ValueError, match="The shapes do not match."):
        merge_cond_shapes(input_)


test_cases = [
    [([(1, 2)], [(3, 4)]), "(1,2)->(3,4)"],
    [([(1, 2), (3, 4)], [(5, 6)]), "(1,2),(3,4)->(5,6)"],
    [([(1, 2)], [(3, 4), (5, 6)]), "(1,2)->(3,4),(5,6)"],
    [([(5,)], [(2,)]), "(5)->(2)"],
    [([()], [()]), "()->()"],
]


@pytest.mark.parametrize(("input_", "expected"), test_cases)
def test_get_ufunc_signature(input_, expected):
    assert _get_ufunc_signature(*input_) == expected



import jax.random as jr
import pytest

from flowjax.bijections import Affine, RationalQuadraticSpline
from flowjax.distributions import StandardNormal
from flowjax.flows import (
    block_neural_autoregressive_flow,
    coupling_flow,
    masked_autoregressive_flow,
    planar_flow,
    triangular_spline_flow,
)

DIM = 3
KEY = jr.PRNGKey(0)
KWARGS = {
    "key": jr.PRNGKey(0),
    "base_dist": StandardNormal((DIM,)),
    "flow_layers": 2,
}


testcases = {
    "BNAF": lambda c: block_neural_autoregressive_flow(**KWARGS, cond_dim=c),
    "triangular_spline_flow": lambda c: triangular_spline_flow(**KWARGS, cond_dim=c),
    "affine_coupling": lambda c: coupling_flow(
        transformer=Affine(), **KWARGS, cond_dim=c
    ),
    "spline_coupling": lambda c: coupling_flow(
        transformer=RationalQuadraticSpline(knots=3, interval=2),
        cond_dim=c,
        **KWARGS,
    ),
    "affine_masked_autoregessive": lambda c: masked_autoregressive_flow(
        transformer=Affine(), **KWARGS, cond_dim=c
    ),
    "planar": lambda c: planar_flow(**KWARGS, cond_dim=c, width_size=5, depth=1),
}


@pytest.mark.parametrize("flowname", testcases.keys())
@pytest.mark.parametrize("cond_dim", [None, 2])
def test_flow_sample(flowname, cond_dim):
    flow = testcases[flowname](cond_dim)
    cond = None if flow.cond_shape is None else jr.normal(KEY, flow.cond_shape)
    try:
        assert flow._sample(KEY, cond).shape == flow.shape
    except NotImplementedError:
        pass


@pytest.mark.parametrize("flowname", testcases.keys())
@pytest.mark.parametrize("cond_dim", [None, 2])
def test_flow_log_prob(flowname, cond_dim):
    flow = testcases[flowname](cond_dim)
    x = jr.normal(KEY, flow.shape)
    cond = None if flow.cond_shape is None else jr.normal(KEY, flow.cond_shape)
    assert flow._log_prob(x, cond).shape == ()



"""Module contains bijections formed by "stacking/concatenating" other bijections."""

from collections.abc import Sequence
from itertools import accumulate

import jax.numpy as jnp
from jaxtyping import Array

from flowjax.bijections.bijection import AbstractBijection
from flowjax.utils import check_shapes_match, merge_cond_shapes


class Concatenate(AbstractBijection):
    """Concatenate bijections along an existing axis, similar to ``jnp.concatenate``.

    See also :class:`Stack`.

    Args:
        bijections: Bijections, to stack into a single bijection.
        axis: Axis along which to stack. Defaults to 0.
    """

    shape: tuple[int, ...]
    cond_shape: tuple[int, ...] | None
    split_idxs: tuple[int, ...]
    bijections: Sequence[AbstractBijection]
    axis: int

    def __init__(self, bijections: Sequence[AbstractBijection], axis: int = 0):
        self.bijections = bijections
        self.axis = axis

        shapes = [b.shape for b in bijections]
        self._argcheck_shapes(shapes)
        axis = range(len(shapes[0]))[axis]  # Avoids issues when axis==-1
        self.shape = (
            shapes[0][:axis] + (sum(s[axis] for s in shapes),) + shapes[0][axis + 1 :]
        )
        self.split_idxs = tuple(accumulate([s[axis] for s in shapes[:-1]]))
        self.cond_shape = merge_cond_shapes([b.cond_shape for b in bijections])

    def transform(self, x, condition=None):
        x_parts = jnp.array_split(x, self.split_idxs, axis=self.axis)
        y_parts = [
            b.transform(x_part, condition)
            for b, x_part in zip(self.bijections, x_parts, strict=True)
        ]
        return jnp.concatenate(y_parts, axis=self.axis)

    def transform_and_log_det(self, x, condition=None):
        x_parts = jnp.array_split(x, self.split_idxs, axis=self.axis)

        ys_log_dets = [
            b.transform_and_log_det(x, condition)
            for b, x in zip(self.bijections, x_parts, strict=True)
        ]

        y_parts, log_dets = zip(*ys_log_dets, strict=True)
        return jnp.concatenate(y_parts, self.axis), sum(log_dets)

    def inverse(self, y, condition=None):
        y_parts = jnp.array_split(y, self.split_idxs, axis=self.axis)
        x_parts = [
            b.inverse(y_part, condition)
            for b, y_part in zip(self.bijections, y_parts, strict=True)
        ]
        return jnp.concatenate(x_parts, axis=self.axis)

    def inverse_and_log_det(self, y, condition=None):
        y_parts = jnp.array_split(y, self.split_idxs, axis=self.axis)

        xs_log_dets = [
            b.inverse_and_log_det(y, condition)
            for b, y in zip(self.bijections, y_parts, strict=True)
        ]

        x_parts, log_dets = zip(*xs_log_dets, strict=True)
        return jnp.concatenate(x_parts, self.axis), sum(log_dets)

    def _argcheck_shapes(self, shapes: Sequence[tuple[int, ...]]):
        axis = range(len(shapes[0]))[self.axis]  # Avoid negative index
        expected_matching = shapes[0][:axis] + shapes[0][axis + 1 :]
        for i, shape in enumerate(shapes):
            if shape[:axis] + shape[axis + 1 :] != expected_matching:
                raise ValueError(
                    f"Expected bijection shapes to match except along axis {axis}, but "
                    f"index 0 had shape {shapes[0]}, and index {i} had shape {shape}.",
                )


class Stack(AbstractBijection):
    """Stack bijections along a new axis (analagous to ``jnp.stack``).

    See also :class:`Concatenate`.

    Args:
        bijections: Bijections.
        axis: Axis along which to stack. Defaults to 0.
    """

    shape: tuple[int, ...]
    cond_shape: tuple[int, ...] | None
    bijections: Sequence[AbstractBijection]
    axis: int

    def __init__(self, bijections: list[AbstractBijection], axis: int = 0):
        self.axis = axis
        self.bijections = bijections

        shapes = [b.shape for b in bijections]
        check_shapes_match(shapes)

        self.shape = shapes[0][:axis] + (len(bijections),) + shapes[0][axis:]
        self.cond_shape = merge_cond_shapes([b.cond_shape for b in bijections])

    def transform(self, x, condition=None):
        x_parts = self._split_and_squeeze(x)
        y_parts = [
            b.transform(x, condition)
            for (b, x) in zip(self.bijections, x_parts, strict=True)
        ]
        return jnp.stack(y_parts, self.axis)

    def transform_and_log_det(self, x, condition=None):
        x_parts = self._split_and_squeeze(x)
        ys_log_det = [
            b.transform_and_log_det(x, condition)
            for b, x in zip(self.bijections, x_parts, strict=True)
        ]

        y_parts, log_dets = zip(*ys_log_det, strict=True)
        return jnp.stack(y_parts, self.axis), sum(log_dets)

    def inverse(self, y, condition=None):
        y_parts = self._split_and_squeeze(y)
        x_parts = [
            b.inverse(y, condition)
            for (b, y) in zip(self.bijections, y_parts, strict=True)
        ]
        return jnp.stack(x_parts, self.axis)

    def inverse_and_log_det(self, y, condition=None):
        y_parts = self._split_and_squeeze(y)
        xs_log_det = [
            b.inverse_and_log_det(y, condition)
            for b, y in zip(self.bijections, y_parts, strict=True)
        ]
        x_parts, log_dets = zip(*xs_log_det, strict=True)
        return jnp.stack(x_parts, self.axis), sum(log_dets)

    def _split_and_squeeze(self, array: Array):
        arrays = jnp.split(array, len(self.bijections), axis=self.axis)
        return (a.squeeze(axis=self.axis) for a in arrays)



"""Distributions, including the abstract and concrete classes."""

import inspect
from abc import abstractmethod
from collections.abc import Callable
from functools import wraps
from math import prod
from typing import ClassVar

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
from equinox import AbstractVar
from jax.nn import log_softmax
from jax.numpy import linalg
from jax.scipy import stats as jstats
from jax.scipy.special import logsumexp
from jax.tree_util import tree_map
from jaxtyping import Array, ArrayLike, PRNGKeyArray, Shaped

from flowjax.bijections import (
    AbstractBijection,
    Affine,
    Chain,
    Exp,
    Scale,
    SoftPlus,
    TriangularAffine,
)
from flowjax.utils import (
    _get_ufunc_signature,
    arraylike_to_array,
    merge_cond_shapes,
)
from flowjax.wrappers import AbstractUnwrappable, BijectionReparam, Lambda, unwrap


class AbstractDistribution(eqx.Module):
    """Abstract distribution class.

    Distributions are registered as jax PyTrees (as they are equinox modules), and as
    such they are compatible with normal jax operations.

    Concrete subclasses can be implemented as follows:
        (1) Inherit from :class:`AbstractDistribution`.
        (2) Define the abstract attributes ``shape`` and ``cond_shape``.
            ``cond_shape`` should be ``None`` for unconditional distributions.
        (3) Define the abstract methods `_sample` and `_log_prob`.

    See the source code for :class:`StandardNormal` for a simple concrete example.

    Attributes:
        shape: Tuple denoting the shape of a single sample from the distribution.
        cond_shape: Tuple denoting the shape of an instance of the conditioning
            variable. This should be None for unconditional distributions.

    """

    shape: AbstractVar[tuple[int, ...]]
    cond_shape: AbstractVar[tuple[int, ...] | None]

    @abstractmethod
    def _log_prob(self, x: Array, condition: Array | None = None) -> Array:
        """Evaluate the log probability of point x.

        This method should be be valid for inputs with shapes matching
        ``distribution.shape`` and ``distribution.cond_shape`` for conditional
        distributions (i.e. the method defined for unbatched inputs).
        """

    @abstractmethod
    def _sample(self, key: PRNGKeyArray, condition: Array | None = None) -> Array:
        """Sample a point from the distribution.

        This method should return a single sample with shape matching
        ``distribution.shape``.
        """

    def _sample_and_log_prob(self, key: PRNGKeyArray, condition: Array | None = None):
        """Sample a point from the distribution, and return its log probability."""
        x = self._sample(key, condition)
        return x, self._log_prob(x, condition)

    def log_prob(self, x: ArrayLike, condition: ArrayLike | None = None) -> Array:
        """Evaluate the log probability.

        Uses numpy-like broadcasting if additional leading dimensions are passed.

        Args:
            x: Points at which to evaluate density.
            condition: Conditioning variables. Defaults to None.

        Returns:
            Array: Jax array of log probabilities.
        """
        self = unwrap(self)
        x = arraylike_to_array(x, err_name="x", dtype=float)
        if self.cond_shape is not None:
            condition = arraylike_to_array(condition, err_name="condition", dtype=float)
        lps = self._vectorize(self._log_prob)(x, condition)
        return jnp.where(jnp.isnan(lps), -jnp.inf, lps)

    def sample(
        self,
        key: PRNGKeyArray,
        sample_shape: tuple[int, ...] = (),
        condition: ArrayLike | None = None,
    ) -> Array:
        """Sample from the distribution.

        For unconditional distributions, the output will be of shape
        ``sample_shape + dist.shape``. For conditional distributions, a batch dimension
        in the condition is supported, and the output shape will be
        ``sample_shape + condition_batch_shape + dist.shape``.
        See the example for more information.

        Args:
            key: Jax random key.
            condition: Conditioning variables. Defaults to None.
            sample_shape: Sample shape. Defaults to ().

        Example:
            The below example shows the behaviour of sampling, for an unconditional
            and a conditional distribution.

            .. testsetup::

                from flowjax.distributions import StandardNormal
                import jax.random as jr
                import jax.numpy as jnp
                from flowjax.flows import coupling_flow
                from flowjax.bijections import Affine
                # For a unconditional distribution:
                key = jr.PRNGKey(0)
                dist = StandardNormal((2,))
                # For a conditional distribution
                cond_dist = coupling_flow(
                    key, base_dist=StandardNormal((2,)), cond_dim=3
                    )

            For an unconditional distribution:

            .. doctest::

                >>> dist.shape
                (2,)
                >>> samples = dist.sample(key, (10, ))
                >>> samples.shape
                (10, 2)

            For a conditional distribution:

            .. doctest::

                >>> cond_dist.shape
                (2,)
                >>> cond_dist.cond_shape
                (3,)
                >>> # Sample 10 times for a particular condition
                >>> samples = cond_dist.sample(key, (10,), condition=jnp.ones(3))
                >>> samples.shape
                (10, 2)
                >>> # Sampling, batching over a condition
                >>> samples = cond_dist.sample(key, condition=jnp.ones((5, 3)))
                >>> samples.shape
                (5, 2)
                >>> # Sample 10 times for each of 5 conditioning variables
                >>> samples = cond_dist.sample(key, (10,), condition=jnp.ones((5, 3)))
                >>> samples.shape
                (10, 5, 2)


        """
        self = unwrap(self)
        if self.cond_shape is not None:
            condition = arraylike_to_array(condition, err_name="condition")
        keys = self._get_sample_keys(key, sample_shape, condition)
        return self._vectorize(self._sample)(keys, condition)

    def sample_and_log_prob(
        self,
        key: PRNGKeyArray,
        sample_shape: tuple[int, ...] = (),
        condition: ArrayLike | None = None,
    ) -> tuple[Array, Array]:
        """Sample the distribution and return the samples with their log probabilities.

        For transformed distributions (especially flows), this will generally be more
        efficient than calling the methods seperately. Refer to the
        :py:meth:`~flowjax.distributions.AbstractDistribution.sample` documentation for
        more information.

        Args:
            key: Jax random key.
            condition: Conditioning variables. Defaults to None.
            sample_shape: Sample shape. Defaults to ().
        """
        self = unwrap(self)
        if self.cond_shape is not None:
            condition = arraylike_to_array(condition, err_name="condition")
        keys = self._get_sample_keys(key, sample_shape, condition)
        return self._vectorize(self._sample_and_log_prob)(keys, condition)

    @property
    def ndim(self) -> int:
        """Number of dimensions in the distribution (the length of the shape)."""
        return len(self.shape)

    @property
    def cond_ndim(self) -> None | int:
        """Number of dimensions of the conditioning variable (length of cond_shape)."""
        return None if self.cond_shape is None else len(self.cond_shape)

    def _vectorize(self, method: Callable) -> Callable:
        """Returns a vectorized version of the distribution method."""
        # Get shapes without broadcasting - note the (2, ) corresponds to key arrays.
        maybe_cond = [] if self.cond_shape is None else [self.cond_shape]
        in_shapes = {
            "_sample_and_log_prob": [(2,)] + maybe_cond,
            "_sample": [(2,)] + maybe_cond,
            "_log_prob": [self.shape] + maybe_cond,
        }
        out_shapes = {
            "_sample_and_log_prob": [self.shape, ()],
            "_sample": [self.shape],
            "_log_prob": [()],
        }
        in_shapes, out_shapes = in_shapes[method.__name__], out_shapes[method.__name__]

        def _check_shapes(method):
            # Wraps unvectorised method with shape checking
            @wraps(method)
            def _wrapper(*args, **kwargs):
                bound = inspect.signature(method).bind(*args, **kwargs)
                for in_shape, (name, arg) in zip(
                    in_shapes,
                    bound.arguments.items(),
                    strict=False,
                ):
                    if arg.shape != in_shape:
                        raise ValueError(
                            f"Expected trailing dimensions matching {in_shape} for "
                            f"{name}; got {arg.shape}.",
                        )
                return method(*args, **kwargs)

            return _wrapper

        signature = _get_ufunc_signature(in_shapes, out_shapes)
        ex = frozenset([1]) if self.cond_shape is None else frozenset()
        return jnp.vectorize(_check_shapes(method), signature=signature, excluded=ex)

    def _get_sample_keys(
        self,
        key: PRNGKeyArray,
        sample_shape: tuple[int, ...],
        condition,
    ):
        if self.cond_ndim is not None:
            leading_cond_shape = condition.shape[: -self.cond_ndim or None]
        else:
            leading_cond_shape = ()
        key_shape = sample_shape + leading_cond_shape
        key_size = max(1, prod(key_shape))  # Still need 1 key for scalar sample
        return jnp.reshape(jr.split(key, key_size), (*key_shape, 2))


class AbstractTransformed(AbstractDistribution):
    """Abstract class respresenting transformed distributions.

    We take the forward bijection for use in sampling, and the inverse for use in
    density evaluation. See also :class:`Transformed`.

    Concete implementations should subclass :class:`AbstractTransformed`, and
    define the abstract attributes ``base_dist`` and ``bijection``. See the source code
    for :class:`Normal` as a simple example.

    Attributes:
        base_dist: The base distribution.
        bijection: The transformation to apply.
    """

    base_dist: AbstractVar[AbstractDistribution]
    bijection: AbstractVar[AbstractBijection]

    def _log_prob(self, x, condition=None):
        z, log_abs_det = self.bijection.inverse_and_log_det(x, condition)
        p_z = self.base_dist._log_prob(z, condition)
        return p_z + log_abs_det

    def _sample(self, key, condition=None):
        base_sample = self.base_dist._sample(key, condition)
        return self.bijection.transform(base_sample, condition)

    def _sample_and_log_prob(
        self,
        key: PRNGKeyArray,
        condition: Array | None = None,
    ):  # TODO add overide decorator when python>=3.12 is common
        # We override to avoid computing the inverse transformation.
        base_sample, log_prob_base = self.base_dist._sample_and_log_prob(key, condition)
        sample, forward_log_dets = self.bijection.transform_and_log_det(
            base_sample,
            condition,
        )
        return sample, log_prob_base - forward_log_dets

    def __check_init__(self):  # TODO test errors and test conditional base distribution
        """Checks cond_shape is compatible in both bijection and distribution."""
        if (
            self.base_dist.cond_shape is not None
            and self.bijection.cond_shape is not None
            and self.base_dist.cond_shape != self.bijection.cond_shape
        ):
            raise ValueError(
                "The base distribution and bijection are both conditional "
                "but have mismatched cond_shape attributes. Base distribution has"
                f"{self.base_dist.cond_shape}, and the bijection has"
                f"{self.bijection.cond_shape}.",
            )

    def merge_transforms(self):
        """Unnests nested transformed distributions.

        Returns an equivilent distribution, but ravelling nested
        :class:`AbstractTransformed` distributions such that the returned distribution
        has a base distribution that is not an :class:`AbstractTransformed` instance.
        """
        if not isinstance(self.base_dist, AbstractTransformed):
            return self
        base_dist = self.base_dist
        bijections = [self.bijection]
        while isinstance(base_dist, AbstractTransformed):
            bijections.append(base_dist.bijection)
            base_dist = base_dist.base_dist
        bijection = Chain(list(reversed(bijections))).merge_chains()
        return Transformed(base_dist, bijection)

    @property
    def shape(self) -> tuple[int, ...]:
        return self.base_dist.shape

    @property
    def cond_shape(self) -> tuple[int, ...] | None:
        return merge_cond_shapes((self.bijection.cond_shape, self.base_dist.cond_shape))


class Transformed(AbstractTransformed):
    """Form a distribution like object using a base distribution and a bijection.

    We take the forward bijection for use in sampling, and the inverse
    bijection for use in density evaluation.

    .. warning::
            It is the currently the users responsibility to ensure the bijection is
            valid across the entire support of the distribution. Failure to do so may
            lead to to unexpected results.

    Args:
        base_dist: Base distribution.
        bijection: Bijection to transform distribution.

    Example:
        .. doctest::

            >>> from flowjax.distributions import StandardNormal, Transformed
            >>> from flowjax.bijections import Affine
            >>> normal = StandardNormal()
            >>> bijection = Affine(1)
            >>> transformed = Transformed(normal, bijection)
    """

    base_dist: AbstractDistribution
    bijection: AbstractBijection


class AbstractLocScaleDistribution(AbstractTransformed):
    """Abstract distribution class for affine transformed distributions."""

    base_dist: AbstractVar[AbstractDistribution]
    bijection: AbstractVar[Affine]

    @property
    def loc(self):
        """Location of the distribution."""
        return self.bijection.loc

    @property
    def scale(self):
        """Scale of the distribution."""
        return unwrap(self.bijection.scale)


class StandardNormal(AbstractDistribution):
    """Standard normal distribution.

    Note unlike :class:`Normal`, this has no trainable parameters.

    Args:
        shape: The shape of the distribution. Defaults to ().
    """

    shape: tuple[int, ...] = ()
    cond_shape: ClassVar[None] = None

    def _log_prob(self, x, condition=None):
        return jstats.norm.logpdf(x).sum()

    def _sample(self, key, condition=None):
        return jr.normal(key, self.shape)


class Normal(AbstractLocScaleDistribution):
    """An independent Normal distribution with mean and std for each dimension.

    ``loc`` and ``scale`` should broadcast to the desired shape of the distribution.

    Args:
        loc: Means. Defaults to 0. Defaults to 0.
        scale: Standard deviations. Defaults to 1.
    """

    base_dist: StandardNormal
    bijection: Affine

    def __init__(self, loc: ArrayLike = 0, scale: ArrayLike = 1):
        self.base_dist = StandardNormal(
            jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale)),
        )
        self.bijection = Affine(loc=loc, scale=scale)


class LogNormal(AbstractTransformed):
    """Log normal distribution.

    ``loc`` and ``scale`` here refers to the underlying normal distribution.

    Args:
        loc: Location paramter. Defaults to 0.
        scale: Scale parameter. Defaults to 1.
    """

    base_dist: StandardNormal
    bijection: Chain

    def __init__(self, loc: ArrayLike = 0, scale: ArrayLike = 1):
        shape = jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
        self.base_dist = StandardNormal(shape)
        self.bijection = Chain([Affine(loc, scale), Exp(shape)])


class MultivariateNormal(AbstractTransformed):
    """Multivariate normal distribution.

    Internally this is parameterised using the Cholesky decomposition of the covariance
    matrix.

    Args:
        loc: The location/mean parameter vector. If this is scalar it is broadcast to
            the dimension implied by the covariance matrix.
        covariance: Covariance matrix.
    """

    base_dist: StandardNormal
    bijection: TriangularAffine

    def __init__(
        self,
        loc: Shaped[ArrayLike, "#dim"],
        covariance: Shaped[Array, "dim dim"],
    ):
        self.bijection = TriangularAffine(loc, linalg.cholesky(covariance))
        self.base_dist = StandardNormal(self.bijection.shape)

    @property
    def loc(self):
        """Location (mean) of the distribution."""
        return self.bijection.loc

    @property
    def covariance(self):
        """The covariance matrix."""
        cholesky = unwrap(self.bijection.triangular)
        return cholesky @ cholesky.T


class _StandardUniform(AbstractDistribution):
    r"""Standard Uniform distribution."""

    shape: tuple[int, ...] = ()
    cond_shape: ClassVar[None] = None

    def _log_prob(self, x, condition=None):
        return jstats.uniform.logpdf(x).sum()

    def _sample(self, key, condition=None):
        return jr.uniform(key, shape=self.shape)


class Uniform(AbstractLocScaleDistribution):
    """Uniform distribution.

    ``minval`` and ``maxval`` should broadcast to the desired distribution shape.

    Args:
        minval: Minimum values.
        maxval: Maximum values.
    """

    base_dist: _StandardUniform
    bijection: Affine

    def __init__(self, minval: ArrayLike, maxval: ArrayLike):
        shape = jnp.broadcast_shapes(jnp.shape(minval), jnp.shape(maxval))
        minval, maxval = eqx.error_if(
            (minval, maxval), maxval <= minval, "minval must be less than the maxval."
        )
        self.base_dist = _StandardUniform(shape)
        self.bijection = Affine(loc=minval, scale=maxval - minval)

    @property
    def minval(self):
        """Minimum value of the uniform distribution."""
        return self.bijection.loc

    @property
    def maxval(self):
        """Maximum value of the uniform distribution."""
        return self.bijection.loc + unwrap(self.bijection.scale)


class _StandardGumbel(AbstractDistribution):
    """Standard gumbel distribution (https://en.wikipedia.org/wiki/Gumbel_distribution)."""

    shape: tuple[int, ...] = ()
    cond_shape: ClassVar[None] = None

    def _log_prob(self, x, condition=None):
        return -(x + jnp.exp(-x)).sum()

    def _sample(self, key, condition=None):
        return jr.gumbel(key, shape=self.shape)


class Gumbel(AbstractLocScaleDistribution):
    """Gumbel distribution (https://en.wikipedia.org/wiki/Gumbel_distribution).

    ``loc`` and ``scale`` should broadcast to the dimension of the distribution.

    Args:
        loc: Location paramter. Defaults to 0.
        scale: Scale parameter. Defaults to 1.
    """

    base_dist: _StandardGumbel
    bijection: Affine

    def __init__(self, loc: ArrayLike = 0, scale: ArrayLike = 1):
        self.base_dist = _StandardGumbel(
            jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale)),
        )
        self.bijection = Affine(loc, scale)


class _StandardCauchy(AbstractDistribution):
    """Implements standard cauchy distribution (loc=0, scale=1).

    Ref: https://en.wikipedia.org/wiki/Cauchy_distribution.
    """

    shape: tuple[int, ...] = ()
    cond_shape: ClassVar[None] = None

    def _log_prob(self, x, condition=None):
        return jstats.cauchy.logpdf(x).sum()

    def _sample(self, key, condition=None):
        return jr.cauchy(key, shape=self.shape)


class Cauchy(AbstractLocScaleDistribution):
    """Cauchy distribution (https://en.wikipedia.org/wiki/Cauchy_distribution).

    ``loc`` and ``scale`` should broadcast to the dimension of the distribution.

    Args:
        loc: Location paramter. Defaults to 0.
        scale: Scale parameter. Defaults to 1.
    """

    base_dist: _StandardCauchy
    bijection: Affine

    def __init__(self, loc: ArrayLike = 0, scale: ArrayLike = 1):
        self.base_dist = _StandardCauchy(
            jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale)),
        )
        self.bijection = Affine(loc, scale)


class _StandardStudentT(AbstractDistribution):
    """Implements student T distribution with specified degrees of freedom."""

    shape: tuple[int, ...]
    cond_shape: ClassVar[None] = None
    df: Array | AbstractUnwrappable[Array]

    def __init__(self, df: ArrayLike):
        df = arraylike_to_array(df, dtype=float)
        df = eqx.error_if(df, df <= 0, "Degrees of freedom values must be positive.")
        self.shape = jnp.shape(df)
        self.df = BijectionReparam(df, SoftPlus())

    def _log_prob(self, x, condition=None):
        return jstats.t.logpdf(x, df=self.df).sum()

    def _sample(self, key, condition=None):
        return jr.t(key, df=self.df, shape=self.shape)


class StudentT(AbstractLocScaleDistribution):
    """Student T distribution (https://en.wikipedia.org/wiki/Student%27s_t-distribution).

    ``df``, ``loc`` and ``scale`` broadcast to the dimension of the distribution.

    Args:
        df: The degrees of freedom.
        loc: Location parameter. Defaults to 0.
        scale: Scale parameter. Defaults to 1.
    """

    base_dist: _StandardStudentT
    bijection: Affine

    def __init__(self, df: ArrayLike, loc: ArrayLike = 0, scale: ArrayLike = 1):
        df, loc, scale = jnp.broadcast_arrays(df, loc, scale)
        self.base_dist = _StandardStudentT(df)
        self.bijection = Affine(loc, scale)

    @property
    def df(self):
        """The degrees of freedom of the distribution."""
        return unwrap(self.base_dist.df)


class _StandardLaplace(AbstractDistribution):
    """Implements standard laplace distribution (loc=0, scale=1)."""

    shape: tuple[int, ...] = ()
    cond_shape: ClassVar[None] = None

    def _log_prob(self, x, condition=None):
        return jstats.laplace.logpdf(x).sum()

    def _sample(self, key, condition=None):
        return jr.laplace(key, shape=self.shape)


class Laplace(AbstractLocScaleDistribution):
    """Laplace distribution.

    ``loc`` and ``scale`` should broadcast to the dimension of the distribution.

    Args:
        loc: Location paramter. Defaults to 0.
        scale: Scale parameter. Defaults to 1.
    """

    base_dist: _StandardLaplace
    bijection: Affine

    def __init__(self, loc: ArrayLike = 0, scale: ArrayLike = 1):
        shape = jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale))
        self.base_dist = _StandardLaplace(shape)
        self.bijection = Affine(loc, scale)


class _StandardExponential(AbstractDistribution):
    shape: tuple[int, ...] = ()
    cond_shape: ClassVar[None] = None

    def _log_prob(self, x, condition=None):
        return jstats.expon.logpdf(x).sum()

    def _sample(self, key, condition=None):
        return jr.exponential(key, shape=self.shape)


class Exponential(AbstractTransformed):
    """Exponential distribution.

    Args:
        rate: The rate parameter (1 / scale).
    """

    base_dist: _StandardExponential
    bijection: Scale

    def __init__(self, rate: ArrayLike = 1):
        self.base_dist = _StandardExponential(jnp.shape(rate))
        self.bijection = Scale(1 / rate)

    @property
    def rate(self):
        return 1 / unwrap(self.bijection.scale)


class _StandardLogistic(AbstractDistribution):
    shape: tuple[int, ...] = ()
    cond_shape: ClassVar[None] = None

    def _sample(self, key, condition=None):
        return jr.logistic(key, self.shape)

    def _log_prob(self, x, condition=None):
        return jstats.logistic.logpdf(x).sum()


class Logistic(AbstractLocScaleDistribution):
    """Logistic distribution.

    ``loc`` and ``scale`` should broadcast to the shape of the distribution.

    Args:
        loc: Location parameter. Defaults to 0.
        scale: Scale parameter. Defaults to 1.
    """

    base_dist: _StandardLogistic
    bijection: Affine

    def __init__(self, loc: ArrayLike = 0, scale: ArrayLike = 1):
        self.base_dist = _StandardLogistic(
            shape=jnp.broadcast_shapes(jnp.shape(loc), jnp.shape(scale)),
        )
        self.bijection = Affine(loc=loc, scale=scale)


class VmapMixture(AbstractDistribution):
    """Create a mixture distribution.

    Given a distribution in which the arrays have a leading dimension with size matching
    the number of components, and a set of weights, create a mixture distribution.

    Args:
        dist: Distribution with a leading dimension in arrays with size equal to the
            number of mixture components. Often it is convenient to construct this with
            with a pattern like ``eqx.filter_vmap(MyDistribution)(my_params)``.
        weights: The positive, but possibly unnormalized component weights.
    """

    shape: tuple[int, ...]
    cond_shape: tuple[int, ...] | None
    log_normalized_weights: Array | AbstractUnwrappable[Array]
    dist: AbstractDistribution

    def __init__(
        self,
        dist: AbstractDistribution,
        weights: ArrayLike,
    ):
        weights = eqx.error_if(weights, weights <= 0, "Weights must be positive.")
        self.dist = dist
        self.log_normalized_weights = Lambda(lambda w: log_softmax(w), jnp.log(weights))
        self.shape = dist.shape
        self.cond_shape = dist.cond_shape

    def _log_prob(self, x, condition=None):
        log_probs = eqx.filter_vmap(lambda d: d._log_prob(x, condition))(self.dist)
        return logsumexp(log_probs + self.log_normalized_weights)

    def _sample(self, key, condition=None):
        key1, key2 = jr.split(key)
        component = jr.categorical(key1, self.log_normalized_weights)
        component_dist = tree_map(
            lambda leaf: leaf[component] if isinstance(leaf, Array) else leaf,
            tree=self.dist,
        )
        return component_dist._sample(key2, condition)



"""Utility functions for training."""

from collections.abc import Callable, Sequence
from functools import partial

import equinox as eqx
import jax.numpy as jnp
import jax.random as jr
import optax
from jax import jit
from jaxtyping import Array, PRNGKeyArray, PyTree, Scalar, Shaped


@eqx.filter_jit
def step(
    params: PyTree,
    static: PyTree,
    *args,
    optimizer: optax.GradientTransformation,
    opt_state: PyTree,
    loss_fn: Callable[[PyTree, PyTree], Scalar],
):
    """Carry out a training step.

    Args:
        params: Parameters for the model
        static: Static components of the model.
        *args: Arguments passed to the loss function.
        optimizer: Optax optimizer.
        opt_state: Optimizer state.
        loss_fn: The loss function. This should take params and static as the first two
            arguments.

    Returns:
        tuple: (params, opt_state, loss_val)
    """
    loss_val, grads = eqx.filter_value_and_grad(loss_fn)(params, static, *args)
    updates, opt_state = optimizer.update(grads, opt_state, params=params)
    params = eqx.apply_updates(params, updates)
    return params, opt_state, loss_val


def train_val_split(
    key: PRNGKeyArray,
    arrays: Sequence[Shaped[Array, "batch ..."]],
    val_prop: float = 0.1,
):
    """Random train validation split for a sequence of arrays.

    Args:
        key: Jax random key.
        arrays: Sequence of arrays, with matching size on axis 0.
        val_prop: Proportion of data to use for validation. Defaults to 0.1.

    Returns:
        A tuple containing the train arrays and the validation arrays.
    """
    if not 0 <= val_prop <= 1:
        raise ValueError("val_prop should be between 0 and 1.")

    num_samples = arrays[0].shape[0]
    if not all(isinstance(a, Shaped[Array, " dim ..."]) for a in arrays):
        raise ValueError("Array dimensions must match along axis 0.")

    n_train = num_samples - round(val_prop * num_samples)
    arrays = [jr.permutation(key, a) for a in arrays]
    train_arrays = [arr[:n_train] for arr in arrays]
    val_arrays = [arr[n_train:] for arr in arrays]
    return train_arrays, val_arrays


@partial(jit, static_argnums=1)
def get_batches(arrays: Sequence[Array], batch_size: int):
    """Reshape a sequence of arrays to have an additional dimension for the batches.

    Specifically, this will transform an array with shape ``(data_len, *rest)`` to
    ``(num_batches, batch_size, *rest)``.

    Considerations:
        - The values in the last batch are dropped if truncated, i.e. if
            ``data_len % batch_size != 0``.
        - If the batch size is greater than the length of the data, then we set the
            batch size to be equal to the data length.

    Args:
        arrays: Sequence of arrays, with shape matching on axis 0.
        batch_size: The batch size.
    """
    data_len = arrays[0].shape[0]
    if not all(arr.shape[0] == data_len for arr in arrays):
        raise ValueError("Array dimensions do not match along the batch axis.")

    return tuple(_add_batch(arr, batch_size) for arr in arrays)


def _add_batch(arr, batch_size):
    """Adds a leading dimension for batches, dropping the last batch if truncated."""
    batch_size = min(batch_size, arr.shape[0])
    n_batches = arr.shape[0] // batch_size
    return arr[: n_batches * batch_size].reshape(n_batches, batch_size, *arr.shape[1:])


def count_fruitless(losses: list[float]) -> int:
    """Count the number of epochs since the minimum loss in a list of losses.

    Args:
        losses: List of losses.

    """
    min_idx = jnp.argmin(jnp.array(losses)).item()
    return len(losses) - min_idx - 1



import jax.numpy as jnp
import pytest

from flowjax.bisection_search import (
    AutoregressiveBisectionInverter,
    _adapt_interval_to_include_root,
    _autoregressive_bisection_search,
    _bisection_search,
)


def target_function(x):
    return x + 4


def test_adapt_interval_to_include_root():
    lower, upper, _ = _adapt_interval_to_include_root(
        target_function,
        lower=jnp.array(1.1),
        upper=jnp.array(1.2),
    )

    assert lower < -4
    assert upper < 1.1  # Upper should improve too

    # If already includes root, shouldn't change anything
    init_lower, init_upper = jnp.array(-10), jnp.array(10)
    lower, upper, iterations = _adapt_interval_to_include_root(
        target_function,
        lower=init_lower,
        upper=init_upper,
    )
    assert lower == init_lower
    assert upper == init_upper
    assert iterations == 0


true_root = -4
adapt_exact_test_cases = [
    (true_root, 10, 0),
    (-10, true_root, 0),
    (-2, 0, 1),  # Lower adapted and hits root
    (-8, -6, 1),  # Upper adapted and hits root
]


@pytest.mark.parametrize(
    ("lower", "upper", "expected_iterations"),
    adapt_exact_test_cases,
)
def test_adapt_interval_to_include_root_exact(lower, upper, expected_iterations):
    # Tests cases where the exact root is found
    lower, upper, iterations = _adapt_interval_to_include_root(
        target_function,
        lower=jnp.array(lower),
        upper=jnp.array(upper),
    )
    assert lower == true_root
    assert upper == true_root
    assert iterations == expected_iterations


def test_bisection_search():
    tol = 0.1
    max_iter = 200

    root, adapt_iterations, iterations = _bisection_search(
        target_function,
        lower=jnp.array(-10),
        upper=jnp.array(10),
        tol=tol,
        max_iter=max_iter,
    )

    assert root == pytest.approx(-4, abs=tol)
    assert iterations < max_iter
    assert adapt_iterations == 0

    # Check max_iter terminates loop
    root, adapt_iterations, iterations = _bisection_search(
        target_function,
        lower=jnp.array(-10),
        upper=jnp.array(10),
        tol=tol,
        max_iter=0,
    )
    assert iterations == 0

    # Check can adapt interval if needed
    root, adapt_iterations, iterations = _bisection_search(
        target_function,
        lower=jnp.array(3),
        upper=jnp.array(4),
        tol=tol,
        max_iter=200,
    )
    assert root == pytest.approx(-4, abs=tol)
    assert adapt_iterations > 0


def test_bisection_search_exact():
    # Tests cases where the exact root is found
    root, _, iterations = _bisection_search(
        target_function,
        lower=jnp.array(true_root - 2),
        upper=jnp.array(true_root + 2),
        tol=0.1,
        max_iter=200,
    )
    assert root == true_root
    assert iterations == 1


def test_autoregressive_bisection_search():
    def autoregressive_func(array):
        return jnp.cumsum(array) + jnp.arange(3)

    tol = 1e-5
    result = _autoregressive_bisection_search(
        autoregressive_fn=autoregressive_func,
        tol=tol,
        lower=jnp.array(-10),
        upper=jnp.array(10),
        length=3,
        max_iter=200,
    )
    assert result == pytest.approx(jnp.array([0, -1, -1]), abs=tol * 3)


def test_autoregressive_bijection_bisection_inverter():
    inverter = AutoregressiveBisectionInverter()

    from flowjax.bijections import Affine

    affine = Affine(jnp.arange(3), 0.2)
    y = jnp.arange(3)
    x_bisection = inverter(affine, y)
    x_true = affine.inverse(y)

    # Note due to accumulation of errors we cannot use tol below.
    assert x_bisection == pytest.approx(x_true, abs=inverter.tol * 3)



"""Rational quadratic spline bijections (https://arxiv.org/abs/1906.04032)."""

from functools import partial
from typing import ClassVar

import jax
import jax.numpy as jnp
from jaxtyping import Array, Float

from flowjax import wrappers
from flowjax.bijections.bijection import AbstractBijection


def _real_to_increasing_on_interval(
    arr: Float[Array, " dim"],
    interval: float | int = 1,
    softmax_adjust: float = 1e-2,
    *,
    pad_with_ends: bool = True,
):
    """Transform unconstrained vector to monotonically increasing positions on [-B, B].

    Args:
        arr: Parameter vector.
        interval: Interval to transform output. Defaults to 1.
        softmax_adjust : Rescales softmax output using
            ``(widths + softmax_adjust/widths.size) / (1 + softmax_adjust)``. e.g.
            0=no adjustment, 1=average softmax output with evenly spaced widths, >1
            promotes more evenly spaced widths.
        pad_with_ends: Whether to pad the with -interval and interval. Defaults to True.
    """
    if softmax_adjust < 0:
        raise ValueError("softmax_adjust should be >= 0.")

    widths = jax.nn.softmax(arr)
    widths = (widths + softmax_adjust / widths.size) / (1 + softmax_adjust)
    widths = widths.at[0].set(widths[0] / 2)
    pos = 2 * interval * jnp.cumsum(widths) - interval

    if pad_with_ends:
        pos = jnp.pad(pos, pad_width=1, constant_values=(-interval, interval))

    return pos


class RationalQuadraticSpline(AbstractBijection):
    """Scalar RationalQuadraticSpline transformation (https://arxiv.org/abs/1906.04032).

    Args:
        knots: Number of knots.
        interval: interval to transform, [-interval, interval].
        min_derivative: Minimum dervivative. Defaults to 1e-3.
        softmax_adjust: Controls minimum bin width and height by rescaling softmax
            output, e.g. 0=no adjustment, 1=average softmax output with evenly spaced
            widths, >1 promotes more evenly spaced widths. See
            ``real_to_increasing_on_interval``. Defaults to 1e-2.
    """

    knots: int
    interval: float | int
    softmax_adjust: float | int
    min_derivative: float
    x_pos: Array | wrappers.AbstractUnwrappable[Array]
    y_pos: Array | wrappers.AbstractUnwrappable[Array]
    derivatives: Array | wrappers.AbstractUnwrappable[Array]
    shape: ClassVar[tuple] = ()
    cond_shape: ClassVar[None] = None

    def __init__(
        self,
        *,
        knots: int,
        interval: float | int,
        min_derivative: float = 1e-3,
        softmax_adjust: float | int = 1e-2,
    ):
        self.knots = knots
        self.interval = interval
        self.softmax_adjust = softmax_adjust
        self.min_derivative = min_derivative

        # Inexact arrays
        pos_parameterization = partial(
            _real_to_increasing_on_interval,
            interval=interval,
            softmax_adjust=softmax_adjust,
        )

        self.x_pos = wrappers.Lambda(pos_parameterization, jnp.zeros(knots))
        self.y_pos = wrappers.Lambda(pos_parameterization, jnp.zeros(knots))
        self.derivatives = wrappers.Lambda(
            lambda arr: jax.nn.softplus(arr) + self.min_derivative,
            jnp.full(knots + 2, jnp.log(jnp.exp(1 - min_derivative) - 1)),
        )

    def transform(self, x, condition=None):
        # Following notation from the paper
        x_pos, y_pos, derivatives = self.x_pos, self.y_pos, self.derivatives
        in_bounds = jnp.logical_and(x >= -self.interval, x <= self.interval)
        x_robust = jnp.where(in_bounds, x, 0)  # To avoid nans
        k = jnp.searchsorted(x_pos, x_robust) - 1  # k is bin number
        xi = (x_robust - x_pos[k]) / (x_pos[k + 1] - x_pos[k])
        sk = (y_pos[k + 1] - y_pos[k]) / (x_pos[k + 1] - x_pos[k])
        dk, dk1, yk, yk1 = derivatives[k], derivatives[k + 1], y_pos[k], y_pos[k + 1]
        num = (yk1 - yk) * (sk * xi**2 + dk * xi * (1 - xi))
        den = sk + (dk1 + dk - 2 * sk) * xi * (1 - xi)
        y = yk + num / den  # eq. 4

        # avoid numerical precision issues transforming from in -> out of bounds
        y = jnp.clip(y, -self.interval, self.interval)
        return jnp.where(in_bounds, y, x)

    def transform_and_log_det(self, x, condition=None):
        y = self.transform(x)
        derivative = self.derivative(x)
        return y, jnp.log(derivative).sum()

    def inverse(self, y, condition=None):
        # Following notation from the paper
        x_pos, y_pos, derivatives = self.x_pos, self.y_pos, self.derivatives
        in_bounds = jnp.logical_and(y >= -self.interval, y <= self.interval)
        y_robust = jnp.where(in_bounds, y, 0)  # To avoid nans
        k = jnp.searchsorted(y_pos, y_robust) - 1
        xk, xk1, yk, yk1 = x_pos[k], x_pos[k + 1], y_pos[k], y_pos[k + 1]
        sk = (yk1 - yk) / (xk1 - xk)
        y_delta_s_term = (y_robust - yk) * (
            derivatives[k + 1] + derivatives[k] - 2 * sk
        )
        a = (yk1 - yk) * (sk - derivatives[k]) + y_delta_s_term
        b = (yk1 - yk) * derivatives[k] - y_delta_s_term
        c = -sk * (y_robust - yk)
        sqrt_term = jnp.sqrt(b**2 - 4 * a * c)
        xi = (2 * c) / (-b - sqrt_term)
        x = xi * (xk1 - xk) + xk

        # avoid numerical precision issues transforming from in -> out of bounds
        x = jnp.clip(x, -self.interval, self.interval)
        return jnp.where(in_bounds, x, y)

    def inverse_and_log_det(self, y, condition=None):
        x = self.inverse(y)
        derivative = self.derivative(x)
        return x, -jnp.log(derivative).sum()

    def derivative(self, x) -> Array:
        """The derivative dy/dx of the forward transformation."""
        # Following notation from the paper (eq. 5)
        x_pos, y_pos, derivatives = self.x_pos, self.y_pos, self.derivatives
        in_bounds = jnp.logical_and(x >= -self.interval, x <= self.interval)
        x_robust = jnp.where(in_bounds, x, 0)  # To avoid nans
        k = jnp.searchsorted(x_pos, x_robust) - 1
        xi = (x_robust - x_pos[k]) / (x_pos[k + 1] - x_pos[k])
        sk = (y_pos[k + 1] - y_pos[k]) / (x_pos[k + 1] - x_pos[k])
        dk, dk1 = derivatives[k], derivatives[k + 1]
        num = sk**2 * (dk1 * xi**2 + 2 * sk * xi * (1 - xi) + dk * (1 - xi) ** 2)
        den = (sk + (dk1 + dk - 2 * sk) * xi * (1 - xi)) ** 2
        derivative = num / den
        return jnp.where(in_bounds, derivative, 1.0)



import jax
import jax.numpy as jnp
import pytest
from jax import random

from flowjax.bijections.block_autoregressive_network import (
    BlockAutoregressiveNetwork,
    block_autoregressive_linear,
)
from flowjax.wrappers import unwrap


def test_block_autoregressive_linear():
    block_shape = (3, 2)
    linear, log_jac_3d_fn = block_autoregressive_linear(
        jax.random.PRNGKey(0),
        n_blocks=3,
        block_shape=block_shape,
    )
    linear = unwrap(linear)  # Applies masking
    log_jac_3d = log_jac_3d_fn(linear)
    assert log_jac_3d.shape == (3, *block_shape)
    assert jnp.all(jnp.isfinite(log_jac_3d))


def test_BlockAutoregressiveNetwork():
    dim = 3
    x = jnp.ones(dim)
    key = random.PRNGKey(0)

    barn = BlockAutoregressiveNetwork(key, dim=dim, cond_dim=None, depth=1, block_dim=4)
    barn = unwrap(barn)
    y = barn.transform(x)
    assert y.shape == (dim,)
    auto_jacobian = jax.jacobian(barn.transform)(x)

    # Check autograd autoregressive
    assert jnp.all(jnp.triu(auto_jacobian, 1) == pytest.approx(0, abs=1e-7))
    assert jnp.all(jnp.diag(auto_jacobian) > 0)


def test_BlockAutoregressiveNetwork_conditioning():
    dim = 3
    cond_dim = 2
    x = jnp.ones(dim)
    key = random.PRNGKey(0)
    barn = BlockAutoregressiveNetwork(
        key,
        dim=dim,
        cond_dim=cond_dim,
        depth=1,
        block_dim=4,
    )
    y1 = barn.transform(x, jnp.ones(cond_dim))
    y2 = barn.transform(x, jnp.zeros(cond_dim))
    assert jnp.all(y1 != y2)



"""Chain bijection which allows sequential application of arbitrary bijections."""

from collections.abc import Sequence

from flowjax.bijections.bijection import AbstractBijection
from flowjax.utils import check_shapes_match, merge_cond_shapes
from flowjax.wrappers import AbstractUnwrappable, unwrap


class Chain(AbstractBijection):
    """Chain together arbitrary bijections to form another bijection.

    Args:
        bijections: Sequence of bijections. The bijection shapes must match, and any
            none None condition shapes must match.
    """

    shape: tuple[int, ...]
    cond_shape: tuple[int, ...] | None
    bijections: tuple[AbstractBijection | AbstractUnwrappable[AbstractBijection], ...]

    def __init__(
        self,
        bijections: Sequence[
            AbstractBijection | AbstractUnwrappable[AbstractBijection]
        ],
    ):
        unwrapped = unwrap(bijections)
        check_shapes_match([b.shape for b in unwrapped])
        self.shape = unwrapped[0].shape
        self.cond_shape = merge_cond_shapes([unwrap(b).cond_shape for b in unwrapped])
        self.bijections = tuple(bijections)

    def transform(self, x, condition=None):
        for bijection in self.bijections:
            x = bijection.transform(x, condition)
        return x

    def transform_and_log_det(self, x, condition=None):
        log_abs_det_jac = 0
        for bijection in self.bijections:
            x, log_abs_det_jac_i = bijection.transform_and_log_det(x, condition)
            log_abs_det_jac += log_abs_det_jac_i.sum()
        return x, log_abs_det_jac

    def inverse(self, y, condition=None):
        for bijection in reversed(self.bijections):
            y = bijection.inverse(y, condition)
        return y

    def inverse_and_log_det(self, y, condition=None):
        log_abs_det_jac = 0
        for bijection in reversed(self.bijections):
            y, log_abs_det_jac_i = bijection.inverse_and_log_det(y, condition)
            log_abs_det_jac += log_abs_det_jac_i.sum()
        return y, log_abs_det_jac

    def __getitem__(self, i: int | slice) -> AbstractBijection:
        if isinstance(i, int):
            return self.bijections[i]
        if isinstance(i, slice):
            return Chain(self.bijections[i])
        raise TypeError(f"Indexing with type {type(i)} is not supported.")

    def __iter__(self):
        yield from self.bijections

    def __len__(self):
        return len(self.bijections)

    def merge_chains(self):
        """Returns an equivilent Chain object, in which nested chains are flattened."""
        bijections = self.bijections
        while any(isinstance(b, Chain) for b in bijections):
            bij = []
            for b in bijections:
                if isinstance(b, Chain):
                    bij.extend(b.bijections)
                else:
                    bij.append(b)
            bijections = bij
        return Chain(bijections)



"""Tanh bijection."""

import math
from typing import ClassVar

import jax.numpy as jnp
from jax.nn import softplus

from flowjax.bijections.bijection import AbstractBijection


def _tanh_log_grad(x):
    # Log gradient vector of tanh transformation.
    return -2 * (x + softplus(-2 * x) - jnp.log(2.0))


class Tanh(AbstractBijection):
    """Tanh bijection."""

    shape: tuple[int, ...] = ()
    cond_shape: ClassVar[None] = None

    def transform(self, x, condition=None):
        return jnp.tanh(x)

    def transform_and_log_det(self, x, condition=None):
        return jnp.tanh(x), jnp.sum(_tanh_log_grad(x))

    def inverse(self, y, condition=None):
        return jnp.arctanh(y)

    def inverse_and_log_det(self, y, condition=None):
        x = jnp.arctanh(y)
        return x, -jnp.sum(_tanh_log_grad(x))


class LeakyTanh(AbstractBijection):
    """Tanh bijection, with a linear transformation beyond +/- max_val.

    The value and gradient of the linear segments are set to match tanh at +/- max_val.
    This bijection can be useful to encourage values to be within an interval, whilst
    avoiding numerical precision issues, or in cases we require a real -> real mapping
    so Tanh is not appropriate.

    Args:
        max_val: Value above or below which the function becomes linear.
        shape: The shape of the bijection. Defaults to ().
    """

    max_val: float
    intercept: float
    linear_grad: float
    shape: tuple[int, ...] = ()
    cond_shape: ClassVar[None] = None

    def __init__(self, max_val: float | int, shape: tuple[int, ...] = ()):
        self.max_val = float(max_val)
        self.linear_grad = math.exp(_tanh_log_grad(max_val))
        self.intercept = math.tanh(max_val) - self.linear_grad * max_val
        self.shape = shape

    def transform(self, x, condition=None):
        is_linear = jnp.abs(x) >= self.max_val
        linear_y = self.linear_grad * x + jnp.sign(x) * self.intercept
        tanh_y = jnp.tanh(x)
        return jnp.where(is_linear, linear_y, tanh_y)

    def transform_and_log_det(self, x, condition=None):
        y = self.transform(x)
        log_grads = jnp.where(
            jnp.abs(x) >= self.max_val,
            jnp.log(self.linear_grad),
            _tanh_log_grad(x),
        )
        return y, jnp.sum(log_grads)

    def inverse(self, y, condition=None):
        is_linear = jnp.abs(y) >= jnp.tanh(self.max_val)
        x_linear = (y - jnp.sign(y) * self.intercept) / self.linear_grad
        x_arctan = jnp.arctanh(y)
        return jnp.where(is_linear, x_linear, x_arctan)

    def inverse_and_log_det(self, y, condition=None):
        x = self.inverse(y)
        log_grads = jnp.where(
            jnp.abs(y) >= jnp.tanh(self.max_val),
            jnp.log(self.linear_grad),
            _tanh_log_grad(x),
        )
        return x, -jnp.sum(log_grads)



""":class:`AbstractUnwrappable` objects and utilities.

These are "placeholder" values for specifying custom behaviour for nodes in a pytree.
Many of these facilitate similar functions to pytorch parameterizations. We use this
for example to apply parameter constraints, masking of parameters etc. To apply the
behaviour, we use :func:`unwrap`, which will replace any :class:`AbstractUnwrappable`
nodes in a pytree with the unwrapped versions.

Unwrapping is automatically called in several places, primarily:

* Prior to calling the bijection methods: ``transform``, ``inverse``,
  ``transform_and_log_det`` and ``inverse_and_log_det``.
* Prior to calling distribution methods: ``log_prob``, ``sample`` and
  ``sample_and_log_prob``.
* Prior to computing the loss functions.

If implementing a custom unwrappable, bear in mind:

* The wrapper should avoid implementing information or logic beyond what is required
  for initialization and unwrapping, as this information will be lost when unwrapping.
* The unwrapping should support broadcasting/vmapped initializations. Otherwise, if
  the unwrappable is created within a batched context, it will fail to unwrap
  correctly.
"""

from abc import abstractmethod
from collections.abc import Callable, Iterable
from typing import TYPE_CHECKING, ClassVar, Generic, TypeVar

import equinox as eqx
import jax.numpy as jnp
from jaxtyping import Int, Scalar

from flowjax.utils import arraylike_to_array

if TYPE_CHECKING:
    from flowjax.bijections import AbstractBijection

import jax
from jax import lax
from jaxtyping import Array, ArrayLike, PyTree

T = TypeVar("T")


def unwrap(tree: PyTree):
    """Unwrap all :class:`AbstractUnwrappable` nodes within a pytree."""
    return jax.tree_util.tree_map(
        f=lambda leaf: (
            leaf.recursive_unwrap() if isinstance(leaf, AbstractUnwrappable) else leaf
        ),
        tree=tree,
        is_leaf=lambda x: isinstance(x, AbstractUnwrappable),
    )


class AbstractUnwrappable(eqx.Module, Generic[T]):
    """An abstract class representing an unwrappable object.

    Unwrappables generally replace nodes in a pytree, in order to specify some custom
    behaviour to apply upon unwrapping before use. This can be used e.g. to apply
    parameter constraints, such as making scale parameters postive, or applying
    stop_gradient before accessing the parameters.

    If ``_dummy`` is set to an array (must have shape ()), this is used for inferring
    vmapped dimensions (and sizes) when calling ``unwrap`` to automatically vecotorize
    the method. In some cases this is important for supporting the case where an
    ``AbstractUnwrappable`` is created within e.g. ``eqx.filter_vmap``.
    """

    _dummy: eqx.AbstractVar[Int[Scalar, ""] | None]

    def recursive_unwrap(self) -> T:
        """Returns the unwrapped pytree, unwrapping subnodes as required."""

        def vectorized_unwrap(unwrappable):
            if unwrappable._dummy is None:
                return unwrappable.unwrap()

            def v_unwrap(unwrappable):
                return unwrappable.unwrap()

            for dim in reversed(unwrappable._dummy.shape):
                v_unwrap = eqx.filter_vmap(v_unwrap, axis_size=dim)
            return v_unwrap(unwrappable)

        flat, tree_def = eqx.tree_flatten_one_level(self)
        tree = jax.tree_util.tree_unflatten(tree_def, unwrap(flat))
        return vectorized_unwrap(tree)

    @abstractmethod
    def unwrap(self) -> T:
        """Returns the unwrapped pytree, assuming no wrapped subnodes exist."""
        pass


class NonTrainable(AbstractUnwrappable[T]):
    """Applies stop gradient to all arraylike leaves before unwrapping.

    See also ``non_trainable``, which is probably a generally prefereable way to achieve
    similar behaviour, which wraps the arraylike leaves directly, rather than the tree.

    Useful to mark pytrees (arrays, submodules, etc) as frozen/non-trainable. We also
    filter out NonTrainable nodes when partitioning parameters for training, or when
    parameterizing bijections in coupling/masked autoregressive flows (transformers).
    """

    tree: T
    _dummy: ClassVar[None] = None

    def unwrap(self) -> T:
        differentiable, static = eqx.partition(self.tree, eqx.is_array_like)
        return eqx.combine(lax.stop_gradient(differentiable), static)


def non_trainable(tree: PyTree):
    """Freezes parameters by wrapping inexact array leaves with ``NonTrainable``.

    Wrapping the arrays rather than the entire tree is often preferable, allowing easier
    access to attributes, compared to wrapping the entire tree.

    Args:
        tree: The pytree.
    """

    def _map_fn(leaf):
        return NonTrainable(leaf) if eqx.is_inexact_array(leaf) else leaf

    return jax.tree_util.tree_map(
        f=_map_fn,
        tree=tree,
        is_leaf=lambda x: isinstance(x, NonTrainable),
    )


def _apply_inverse_and_check_valid(bijection, arr):
    param_inv = bijection._vectorize.inverse(arr)
    return eqx.error_if(
        param_inv,
        jnp.logical_and(jnp.isfinite(arr), ~jnp.isfinite(param_inv)),
        "Non-finite value(s) introduced when reparameterizing. This suggests "
        "the parameter vector passed to BijectionReparam was incompatible with "
        f"the bijection used for reparmeterizing ({type(bijection).__name__}).",
    )


class BijectionReparam(AbstractUnwrappable[Array]):
    """Reparameterize a parameter using a bijection.

    When applying unwrap, ``bijection.transform`` is applied. By default, the inverse
    of the bijection is applied when setting the parameter values.

    Args:
        arr: The parameter to reparameterize. If invert_on_init is False, then this can
            be a ``AbstractUnwrappable[Array]``.
        bijection: A bijection whose shape is broadcastable to ``jnp.shape(arr)``.
        invert_on_init: Whether to apply the inverse transformation when initializing.
            Defaults to True.
    """

    arr: Array | AbstractUnwrappable[Array]
    bijection: "AbstractBijection"
    _dummy: Int[Scalar, ""]

    def __init__(
        self,
        arr: Array | AbstractUnwrappable[Array],
        bijection: "AbstractBijection",
        *,
        invert_on_init: bool = True,
    ):
        if invert_on_init:
            self.arr = _apply_inverse_and_check_valid(bijection, arr)
        else:
            if not isinstance(arr, AbstractUnwrappable):
                arr = arraylike_to_array(arr)
            self.arr = arr
        self.bijection = bijection
        self._dummy = jnp.empty((), int)

    def unwrap(self) -> Array:
        return self.bijection._vectorize.transform(self.arr)


class Where(AbstractUnwrappable[Array]):
    """Applies jnp.where upon unwrapping.

    This can be used to construct masks by setting ``cond=mask`` and ``if_false=0``.
    """

    cond: ArrayLike
    if_true: ArrayLike | AbstractUnwrappable[Array]
    if_false: ArrayLike | AbstractUnwrappable[Array]
    _dummy: ClassVar[None] = None

    def unwrap(self):
        return jnp.where(self.cond, self.if_true, self.if_false)


class WeightNormalization(AbstractUnwrappable[Array]):
    """Applies weight normalization (https://arxiv.org/abs/1602.07868).

    Args:
        weight: The (possibly wrapped) weight matrix.
    """

    weight: Array | AbstractUnwrappable[Array]
    scale: Array | AbstractUnwrappable[Array] = eqx.field(init=False)
    _dummy: ClassVar[None] = None

    def __init__(self, weight: Array | AbstractUnwrappable[Array]):
        from flowjax.bijections import SoftPlus  # Delayed to avoid circular import...

        self.weight = weight
        scale_init = 1 / jnp.linalg.norm(unwrap(weight), axis=-1, keepdims=True)
        self.scale = BijectionReparam(scale_init, SoftPlus())

    def unwrap(self) -> Array:
        weight_norms = jnp.linalg.norm(self.weight, axis=-1, keepdims=True)
        return self.scale * self.weight / weight_norms


class Lambda(AbstractUnwrappable[T]):
    """Unwrap an object by calling fn with (possibly trainable) args and kwargs.

    If the Lambda is created within ``eqx.filter_vmap``, unwrapping is automatically
    vectorized correctly, as long as the vmapped constructor adds leading batch
    dimensions to all arrays in Lambda (the default for ``eqx.filter_vmap``).

    Args:
        fn: Function to call with args, and kwargs.
        *args: Positional arguments to pass to fn.
        **kwargs: Keyword arguments to pass to fn.
    """

    fn: Callable[..., T]
    args: Iterable
    kwargs: dict
    _dummy: Int[Scalar, ""]

    def __init__(self, fn, *args, **kwargs):
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self._dummy = jnp.empty((), int)

    def unwrap(self) -> T:
        return self.fn(*self.args, **self.kwargs)



from functools import partial
from typing import ClassVar

import equinox as eqx
import jax
import jax.numpy as jnp
import jax.random as jr
import numpyro
import numpyro.distributions as ndist
import pytest
from equinox.nn import Linear
from jax.flatten_util import ravel_pytree
from jaxtyping import Array
from numpyro import handlers
from numpyro.distributions.transforms import AffineTransform
from numpyro.infer import MCMC, NUTS, SVI, Trace_ELBO
from numpyro.optim import Adam

from flowjax.bijections import AbstractBijection, AdditiveCondition, Affine, Invert
from flowjax.distributions import (
    LogNormal,
    Normal,
    StandardNormal,
    Transformed,
)
from flowjax.experimental.numpyro import (
    _BijectionToNumpyro,
    _get_batch_shape,
    distribution_to_numpyro,
    register_params,
    sample,
)
from flowjax.flows import block_neural_autoregressive_flow, masked_autoregressive_flow

true_mean, true_std = jnp.ones(2), 2 * jnp.ones(2)


def numpyro_model():
    sample("x", Normal(true_mean, true_std))


def test_get_batch_shape():
    assert _get_batch_shape(jnp.ones((2, 3)), ()) == (2, 3)
    assert _get_batch_shape(jnp.ones(3), (3,)) == ()
    assert _get_batch_shape(jnp.ones((1, 2, 3)), (2, 3)) == (1,)


def test_mcmc():
    "Check that flowjax distributions function in MCMC."
    key = jr.PRNGKey(0)
    mcmc = MCMC(NUTS(numpyro_model), num_warmup=50, num_samples=500)  # 2d N(1, 2)
    key, subkey = jr.split(key)
    mcmc.run(subkey)
    samps = mcmc.get_samples()["x"]
    assert pytest.approx(samps.mean(axis=0), abs=0.2) == true_mean
    assert pytest.approx(samps.std(axis=0), abs=0.2) == true_std

    def plate_model():
        # Note, in flowjax we do not have the concept of batch shape for simplicity.
        # We could add support in the numpyro context later if we wish. Note below,
        # the plate dim is -1 (as the flowjax normal has event dim (2,)).
        with numpyro.plate("obs", 10, dim=-1):
            sample("x", Normal(true_mean, true_std))

    mcmc = MCMC(NUTS(plate_model), num_warmup=50, num_samples=500)  # 2d N(1, 2)
    key, subkey = jr.split(key)
    mcmc.run(subkey)

    samps = mcmc.get_samples()["x"]
    assert samps.shape == (500, 10, 2)
    assert pytest.approx(samps.mean(axis=[0, 1]), abs=0.2) == true_mean
    assert pytest.approx(samps.std(axis=[0, 1]), abs=0.2) == true_std


def test_vi():
    "Check that flowjax distributions can be used as guide/variational distributions."

    def guide(dist):
        dist = register_params("guide", dist)
        sample("x", dist)

    optimizer = Adam(step_size=0.01)

    guide_dist = Normal(jnp.zeros(2), 1)  # 2d N(0, 1)

    svi = SVI(numpyro_model, partial(guide, guide_dist), optimizer, loss=Trace_ELBO())
    svi_result = svi.run(jr.PRNGKey(0), num_steps=1000)

    guide_dist = eqx.combine(svi_result.params["guide"], guide_dist)

    assert pytest.approx(guide_dist.loc, abs=0.2) == true_mean
    assert pytest.approx(guide_dist.scale, abs=0.2) == true_std

    # Test intermediates are used - note BNAF has no inverse so intermediates
    # are required to compute the log_prob in VI.
    guide_dist = block_neural_autoregressive_flow(
        key,
        base_dist=Normal(jnp.zeros(2)),
        invert=False,
        nn_block_dim=1,
    )
    svi = SVI(numpyro_model, partial(guide, guide_dist), optimizer, loss=Trace_ELBO())
    svi_result = svi.run(jr.PRNGKey(0), num_steps=2)  # Check runs


def test_conditional_vi():
    # Similar to amortized variational inference
    dim = 2
    cond_dim = 3

    key, subkey = jr.split(jr.PRNGKey(0))

    true_dist, guide_dist = get_conditional_true_guide(subkey, dim, cond_dim)

    def model():
        cond = sample("cond", ndist.Normal(jnp.zeros((3,))))
        sample("x", true_dist, condition=cond)

    def guide(guide_dist):
        guide_dist = register_params("guide", guide_dist)
        cond = sample("cond", ndist.Normal(jnp.zeros((3,))))
        sample("x", guide_dist, condition=cond)

    optimizer = Adam(step_size=0.01)
    svi = SVI(model, partial(guide, guide_dist), optimizer, loss=Trace_ELBO())

    key, subkey = jr.split(key)
    svi_result = svi.run(subkey, num_steps=1000)

    trained_guide = eqx.combine(svi_result.params["guide"], guide_dist)
    true_params = ravel_pytree(true_dist)[0]
    norm_from_true_init = jnp.linalg.norm(true_params - ravel_pytree(guide_dist)[0])
    norm_from_true_final = jnp.linalg.norm(true_params - ravel_pytree(trained_guide)[0])

    # Arbitrarily, we check the l2 norm between the trained parameters and the truth
    # has more than halved over the course of training
    assert norm_from_true_final < 0.5 * norm_from_true_init


def test_vi_plate():
    plate_dim = 10

    def model():
        with numpyro.plate("obs", plate_dim):
            sample("x", Normal(true_mean, true_std))

    guide_dist = Normal(jnp.ones_like(true_mean), jnp.ones_like(true_std))

    def guide(guide_dist):
        guide = register_params("guide", guide_dist)
        with numpyro.plate("obs", plate_dim):
            sample("x", guide)

    guide = partial(guide, guide_dist)
    optimizer = Adam(step_size=0.01)
    svi = SVI(model, guide, optimizer, loss=Trace_ELBO())

    svi_result = svi.run(jr.PRNGKey(0), num_steps=10000)

    guide_dist = eqx.combine(svi_result.params["guide"], guide_dist)

    assert pytest.approx(guide_dist.loc, abs=0.2) == true_mean
    assert pytest.approx(guide_dist.scale, abs=0.2) == true_std


key = jr.PRNGKey(0)
test_cases = {
    "distribution": [
        StandardNormal(()),
        StandardNormal((1, 2)),
        Normal(),
        Normal(jnp.ones((1, 2))),
    ],
    "sample_shape": [(), (3, 4)],
}


@pytest.mark.parametrize("distribution", test_cases["distribution"])
@pytest.mark.parametrize("sample_shape", test_cases["sample_shape"])
def test_distribution_to_numpyro(distribution, sample_shape):
    wrapped = distribution_to_numpyro(distribution)

    # Same key should lead to same samples
    x = distribution.sample(key, sample_shape)
    assert wrapped.sample(key, sample_shape) == pytest.approx(x)

    # Should give same log prob
    assert wrapped.log_prob(x) == pytest.approx(distribution.log_prob(x))


def test_batched_condition():
    dim = 2
    cond_dim = 3
    key, subkey = jr.split(jr.PRNGKey(0))
    dist = block_neural_autoregressive_flow(
        key,
        base_dist=Normal(jnp.zeros(dim)),
        cond_dim=cond_dim,
    )

    condition = jnp.ones((10, 3))
    wrapped = distribution_to_numpyro(dist, condition)
    assert wrapped.batch_shape == (10,)

    key, subkey = jr.split(key)
    true_dist, guide_dist = get_conditional_true_guide(subkey, dim, cond_dim)

    def model():
        with numpyro.plate("N", 10, dim=-2):
            cond = sample("cond", ndist.Normal(jnp.zeros(3)).to_event())
            sample("x", true_dist, condition=cond)

    def guide(guide_dist):
        guide_dist = register_params("guide", guide_dist)
        with numpyro.plate("N", 10, dim=-2):
            cond = sample("cond", ndist.Normal(jnp.zeros(3)).to_event())
            sample("x", guide_dist, condition=cond)

    optimizer = Adam(step_size=0.01)
    svi = SVI(model, partial(guide, guide_dist), optimizer, loss=Trace_ELBO())
    key, subkey = jr.split(key)
    svi_result = svi.run(subkey, num_steps=1000)

    trained_guide = eqx.combine(svi_result.params["guide"], guide_dist)
    true_params = ravel_pytree(true_dist)[0]
    norm_from_true_init = jnp.linalg.norm(true_params - ravel_pytree(guide_dist)[0])
    norm_from_true_final = jnp.linalg.norm(true_params - ravel_pytree(trained_guide)[0])

    # Arbitrarily, we check the l2 norm between the trained parameters and the truth
    # has more than halved over the course of training
    assert norm_from_true_final < 0.5 * norm_from_true_init


def get_conditional_true_guide(key, dim, cond_dim):
    true_dist, guide_dist = (
        Transformed(
            StandardNormal((dim,)),
            AdditiveCondition(Linear(cond_dim, dim, key=k), (dim,), (cond_dim,)),
        )
        for k in jr.split(key)
    )
    return true_dist, guide_dist


def test_callable_params():
    class MyParams(eqx.Module):
        a: Array
        b: Array

        def __call__(self, *args, **kwargs):
            raise NotImplementedError()

    def model(params):
        params = register_params("my_params", params)
        sample("x", Normal(params.a, params.b))

    my_params = MyParams(jnp.ones(3), jnp.ones(3))
    model = handlers.seed(model, key)
    model_trace = handlers.trace(model).get_trace(my_params)

    assert "my_params" in model_trace
    assert eqx.tree_equal(my_params, model_trace["my_params"]["value"])


def test_expected_elbo():
    dim = 2
    cond_dim = 3
    n_obs = 5
    key, subkey = jr.split(jr.PRNGKey(0))

    likelihood = block_neural_autoregressive_flow(
        key,
        base_dist=Normal(jnp.zeros(dim)),
        cond_dim=cond_dim,
        nn_block_dim=2,
    )
    prior = StandardNormal((cond_dim,))

    key, subkey = jr.split(jr.PRNGKey(0))
    posterior = block_neural_autoregressive_flow(
        key,
        base_dist=Normal(jnp.zeros(cond_dim)),
        cond_dim=dim,
        nn_block_dim=2,
        invert=False,
    )

    def model(obs):
        with numpyro.plate("obs", n_obs):
            theta = sample("theta", prior)
            sample("x", likelihood, obs=obs, condition=theta)

    def guide(obs):
        with numpyro.plate("obs", n_obs):
            sample("theta", posterior, condition=obs)

    def manual_elbo(key):
        guide_samps, guide_lps = posterior.sample_and_log_prob(key, condition=obs)
        model_lps = prior.log_prob(guide_samps) + likelihood.log_prob(obs, guide_samps)
        return model_lps.sum() - guide_lps.sum()

    from numpyro.infer import Trace_ELBO

    key, subkey = jr.split(key)
    obs = jr.normal(subkey, (n_obs, dim))

    key, subkey = jr.split(key)
    elbo = -Trace_ELBO(num_particles=500).loss(subkey, {}, model, guide, obs)

    # Seed handling is different, so we compare stochastic elbo for simplicity
    manual = jax.vmap(manual_elbo)(jr.split(subkey, 500)).mean()
    assert elbo == pytest.approx(manual, abs=1)


def test_conditional_sample_and_log_prob_with_intermediates():

    cond_flow = masked_autoregressive_flow(
        jr.PRNGKey(0),
        base_dist=Normal(jnp.ones(3)),
        cond_dim=2,
        flow_layers=2,
    )
    condition = jnp.linspace(-1, 1, num=20).reshape((10, 2))
    dist = distribution_to_numpyro(cond_flow, condition=condition)

    assert dist.batch_shape == (condition.shape[0],)
    assert dist.base_dist.batch_shape == (condition.shape[0],)

    sample, intermediates = dist.sample_with_intermediates(key)
    z, log_det = intermediates[0]
    assert sample.shape == z.shape
    assert log_det.shape == (condition.shape[0],)
    assert not pytest.approx(sample[0]) == sample[1]  # Ensures different base sample

    expected = cond_flow.log_prob(sample, condition)
    log_probs = dist.log_prob(sample, intermediates)
    assert pytest.approx(expected) == log_probs


def test_BijectionToNumpyro_unconditional():
    loc, scale = jnp.arange(3), jnp.arange(1, 4)
    affine = _BijectionToNumpyro(Affine(loc, scale))
    npro_affine = AffineTransform(
        loc,
        scale,
    )

    x = jnp.array([2, 1, 3])
    assert pytest.approx(affine(x)) == npro_affine(x)
    assert pytest.approx(affine.inv(x)) == npro_affine.inv(x)
    assert (
        pytest.approx(affine.log_abs_det_jacobian(x, y=None))
        == npro_affine.log_abs_det_jacobian(x, y=None).sum()
    )

    transformed, npro_transformed = (
        numpyro.distributions.TransformedDistribution(
            numpyro.distributions.Normal(jnp.zeros(3)),
            aff,
        )
        for aff in [affine, npro_affine]
    )
    expected = npro_transformed.log_prob(jnp.arange(9).reshape(3, 3)).sum(axis=-1)
    observed = transformed.log_prob(jnp.arange(9).reshape(3, 3))
    assert pytest.approx(expected) == observed


def test_invalid_domains_BijectionToNumpyro():
    affine = Affine(jnp.arange(3), jnp.arange(1, 4))

    with pytest.raises(ValueError, match="domain.event_dim"):
        _BijectionToNumpyro(affine, domain=numpyro.distributions.constraints.real)

    with pytest.raises(ValueError, match="codomain.event_dim"):
        _BijectionToNumpyro(affine, codomain=numpyro.distributions.constraints.real)


def test_transformed_reparam():
    loc, scale = 2, 3

    config = {"x": numpyro.infer.reparam.TransformReparam()}
    log_norm = LogNormal(loc, scale)

    def model():
        with numpyro.handlers.reparam(config=config):
            sample("x", log_norm)

    trace = handlers.trace(
        handlers.seed(model, jr.PRNGKey(0)),
    ).get_trace()

    assert "x_base" in trace.keys()
    expected_x = log_norm.bijection.transform(trace["x_base"]["value"])
    assert pytest.approx(expected_x) == trace["x"]["value"]


class _ForwardOnly(AbstractBijection):
    shape: tuple[int, ...] = ()
    cond_shape: ClassVar[None] = None

    def transform(self, x, condition=None):
        return x

    def transform_and_log_det(self, x, condition=None):
        return x, jnp.zeros(())

    def inverse(self, y, condition=None):
        raise NotImplementedError()

    def inverse_and_log_det(self, y, condition=None):
        raise NotImplementedError()


def test_sampling_forward_only():
    dist = Transformed(
        StandardNormal(),
        _ForwardOnly(),
    )
    dist = distribution_to_numpyro(dist)
    dist.sample(jr.PRNGKey(0))


def test_log_prob_inverse_only():
    dist = Transformed(
        StandardNormal(),
        Invert(_ForwardOnly()),
    )
    dist = distribution_to_numpyro(dist)
    dist.log_prob(0)



"""SoftPlus bijection."""

from typing import ClassVar

import jax.numpy as jnp
from jax.nn import softplus

from flowjax.bijections.bijection import AbstractBijection


class SoftPlus(AbstractBijection):
    r"""Transforms to positive domain using softplus :math:`y = \log(1 + \exp(x))`."""

    shape: tuple[int, ...] = ()
    cond_shape: ClassVar[None] = None

    def transform(self, x, condition=None):
        return softplus(x)

    def transform_and_log_det(self, x, condition=None):
        return softplus(x), -softplus(-x).sum()

    def inverse(self, y, condition=None):
        return jnp.log(-jnp.expm1(-y)) + y

    def inverse_and_log_det(self, y, condition=None):
        x = self.inverse(y)
        return x, softplus(-x).sum()
